#!/usr/bin/env python
#
# deepzoom_server - Example web application for serving whole-slide images
#
# Copyright (c) 2010-2015 Carnegie Mellon University
# Copyright (c) 2023      Benjamin Gilbert
#
# This library is free software; you can redistribute it and/or modify it
# under the terms of version 2.1 of the GNU Lesser General Public License
# as published by the Free Software Foundation.
#
# This library is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public
# License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this library.  If not, see <https://www.gnu.org/licenses/>.
#

from __future__ import annotations

from argparse import ArgumentParser
import base64
from collections.abc import Callable, Mapping
from io import BytesIO
import os
from pathlib import Path
import re
from typing import TYPE_CHECKING, Any, Literal
from unicodedata import normalize
import zlib

from PIL import Image, ImageCms
from flask import Flask, Response, abort, make_response, render_template, url_for

if TYPE_CHECKING:
    # Python 3.10+
    from typing import TypeAlias

if os.name == 'nt':
    _dll_path = os.getenv('OPENSLIDE_PATH')
    if _dll_path is not None:
        with os.add_dll_directory(_dll_path):  # type: ignore[attr-defined,unused-ignore]  # noqa: E501
            import openslide
    else:
        import openslide
else:
    import openslide

from openslide import AbstractSlide, ImageSlide, open_slide
from openslide.deepzoom import DeepZoomGenerator

SLIDE_NAME = 'slide'

# Optimized sRGB v2 profile, CC0-1.0 license
# https://github.com/saucecontrol/Compact-ICC-Profiles/blob/bdd84663/profiles/sRGB-v2-micro.icc
# ImageCms.createProfile() generates a v4 profile and Firefox has problems
# with those: https://littlecms.com/blog/2020/09/09/browser-check/
SRGB_PROFILE_BYTES = zlib.decompress(
    base64.b64decode(
        'eNpjYGA8kZOcW8wkwMCQm1dSFOTupBARGaXA/oiBmUGEgZOBj0E2Mbm4wDfYLYQBCIoT'
        'y4uTS4pyGFDAt2sMjCD6sm5GYl7K3IkMtg4NG2wdSnQa5y1V6mPADzhTUouTgfQHII5P'
        'LigqYWBg5AGyecpLCkBsCSBbpAjoKCBbB8ROh7AdQOwkCDsErCYkyBnIzgCyE9KR2ElI'
        'bKhdIMBaCvQsskNKUitKQLSzswEDKAwgop9DwH5jFDuJEMtfwMBg8YmBgbkfIZY0jYFh'
        'eycDg8QthJgKUB1/KwPDtiPJpUVlUGu0gLiG4QfjHKZS5maWk2x+HEJcEjxJfF8Ez4t8'
        'k8iS0VNwVlmjmaVXZ/zacrP9NbdwX7OQshjxFNmcttKwut4OnUlmc1Yv79l0e9/MU8ev'
        'pz4p//jz/38AR4Nk5Q=='
    )
)
SRGB_PROFILE = ImageCms.getOpenProfile(BytesIO(SRGB_PROFILE_BYTES))

if TYPE_CHECKING:
    ColorMode: TypeAlias = Literal[
        'default',
        'absolute-colorimetric',
        'perceptual',
        'relative-colorimetric',
        'saturation',
        'embed',
        'ignore',
    ]
    Transform: TypeAlias = Callable[[Image.Image], None]


class DeepZoomServer(Flask):
    slides: dict[str, DeepZoomGenerator]
    transforms: dict[str, Transform]
    slide_properties: Mapping[str, str]
    associated_images: list[str]
    slide_mpp: float


def create_app(
    config: dict[str, Any] | None = None,
    config_file: Path | None = None,
) -> Flask:
    # Create and configure app
    app = DeepZoomServer(__name__)
    app.config.from_mapping(
        DEEPZOOM_SLIDE=None,
        DEEPZOOM_FORMAT='jpeg',
        DEEPZOOM_TILE_SIZE=254,
        DEEPZOOM_OVERLAP=1,
        DEEPZOOM_LIMIT_BOUNDS=True,
        DEEPZOOM_TILE_QUALITY=75,
        DEEPZOOM_COLOR_MODE='default',
    )
    app.config.from_envvar('DEEPZOOM_TILER_SETTINGS', silent=True)
    if config_file is not None:
        app.config.from_pyfile(config_file)
    if config is not None:
        app.config.from_mapping(config)

    # Open slide
    if app.config['DEEPZOOM_SLIDE'] is None:
        raise ValueError('No slide file specified')
    slidefile = Path(app.config['DEEPZOOM_SLIDE'])
    config_map = {
        'DEEPZOOM_TILE_SIZE': 'tile_size',
        'DEEPZOOM_OVERLAP': 'overlap',
        'DEEPZOOM_LIMIT_BOUNDS': 'limit_bounds',
    }
    opts = {v: app.config[k] for k, v in config_map.items()}
    slide = open_slide(slidefile)
    app.slides = {SLIDE_NAME: DeepZoomGenerator(slide, **opts)}
    app.transforms = {
        SLIDE_NAME: get_transform(slide, app.config['DEEPZOOM_COLOR_MODE'])
    }
    app.associated_images = []
    app.slide_properties = slide.properties
    for name, image in slide.associated_images.items():
        app.associated_images.append(name)
        slug = slugify(name)
        image_slide = ImageSlide(image)
        app.slides[slug] = DeepZoomGenerator(image_slide, **opts)
        app.transforms[slug] = get_transform(
            image_slide, app.config['DEEPZOOM_COLOR_MODE']
        )
    try:
        mpp_x = slide.properties[openslide.PROPERTY_NAME_MPP_X]
        mpp_y = slide.properties[openslide.PROPERTY_NAME_MPP_Y]
        app.slide_mpp = (float(mpp_x) + float(mpp_y)) / 2
    except (KeyError, ValueError):
        app.slide_mpp = 0

    # Set up routes
    @app.route('/')
    def index() -> str:
        slide_url = url_for('dzi', slug=SLIDE_NAME)
        associated_urls = {
            name: url_for('dzi', slug=slugify(name)) for name in app.associated_images
        }
        return render_template(
            'slide-multipane.html',
            slide_url=slide_url,
            associated=associated_urls,
            properties=app.slide_properties,
            slide_mpp=app.slide_mpp,
        )

    @app.route('/<slug>.dzi')
    def dzi(slug: str) -> Response:
        format = app.config['DEEPZOOM_FORMAT']
        try:
            resp = make_response(app.slides[slug].get_dzi(format))
            resp.mimetype = 'application/xml'
            return resp
        except KeyError:
            # Unknown slug
            abort(404)

    @app.route('/<slug>_files/<int:level>/<int:col>_<int:row>.<format>')
    def tile(slug: str, level: int, col: int, row: int, format: str) -> Response:
        format = format.lower()
        if format != 'jpeg' and format != 'png':
            # Not supported by Deep Zoom
            abort(404)
        try:
            tile = app.slides[slug].get_tile(level, (col, row))
        except KeyError:
            # Unknown slug
            abort(404)
        except ValueError:
            # Invalid level or coordinates
            abort(404)
        app.transforms[slug](tile)
        buf = BytesIO()
        tile.save(
            buf,
            format,
            quality=app.config['DEEPZOOM_TILE_QUALITY'],
            icc_profile=tile.info.get('icc_profile'),
        )
        resp = make_response(buf.getvalue())
        resp.mimetype = 'image/%s' % format
        return resp

    return app


def slugify(text: str) -> str:
    text = normalize('NFKD', text.lower()).encode('ascii', 'ignore').decode()
    return re.sub('[^a-z0-9]+', '-', text)


def get_transform(image: AbstractSlide, mode: ColorMode) -> Transform:
    if image.color_profile is None:
        return lambda img: None
    if mode == 'ignore':
        # drop ICC profile from tiles
        return lambda img: img.info.pop('icc_profile')
    elif mode == 'embed':
        # embed ICC profile in tiles
        return lambda img: None
    elif mode == 'default':
        intent = ImageCms.Intent(ImageCms.getDefaultIntent(image.color_profile))
    elif mode == 'absolute-colorimetric':
        intent = ImageCms.Intent.ABSOLUTE_COLORIMETRIC
    elif mode == 'relative-colorimetric':
        intent = ImageCms.Intent.RELATIVE_COLORIMETRIC
    elif mode == 'perceptual':
        intent = ImageCms.Intent.PERCEPTUAL
    elif mode == 'saturation':
        intent = ImageCms.Intent.SATURATION
    else:
        raise ValueError(f'Unknown color mode {mode}')
    transform = ImageCms.buildTransform(
        image.color_profile,
        SRGB_PROFILE,
        'RGB',
        'RGB',
        intent,
        ImageCms.Flags(0),
    )

    def xfrm(img: Image.Image) -> None:
        ImageCms.applyTransform(img, transform, True)
        # Some browsers assume we intend the display's color space if we don't
        # embed the profile.  Pillow's serialization is larger, so use ours.
        img.info['icc_profile'] = SRGB_PROFILE_BYTES

    return xfrm


if __name__ == '__main__':
    parser = ArgumentParser(usage='%(prog)s [options] [SLIDE]')
    parser.add_argument(
        '-B',
        '--ignore-bounds',
        dest='DEEPZOOM_LIMIT_BOUNDS',
        default=True,
        action='store_false',
        help='display entire scan area',
    )
    parser.add_argument(
        '--color-mode',
        dest='DEEPZOOM_COLOR_MODE',
        choices=[
            'default',
            'absolute-colorimetric',
            'perceptual',
            'relative-colorimetric',
            'saturation',
            'embed',
            'ignore',
        ],
        default='default',
        help=(
            'convert tiles to sRGB using default rendering intent of ICC '
            'profile, or specified rendering intent; or embed original '
            'ICC profile; or ignore ICC profile (compat) [default]'
        ),
    )
    parser.add_argument(
        '-c', '--config', metavar='FILE', type=Path, dest='config', help='config file'
    )
    parser.add_argument(
        '-d',
        '--debug',
        dest='DEBUG',
        action='store_true',
        help='run in debugging mode (insecure)',
    )
    parser.add_argument(
        '-e',
        '--overlap',
        metavar='PIXELS',
        dest='DEEPZOOM_OVERLAP',
        type=int,
        help='overlap of adjacent tiles [1]',
    )
    parser.add_argument(
        '-f',
        '--format',
        dest='DEEPZOOM_FORMAT',
        choices=['jpeg', 'png'],
        help='image format for tiles [jpeg]',
    )
    parser.add_argument(
        '-l',
        '--listen',
        metavar='ADDRESS',
        dest='host',
        default='127.0.0.1',
        help='address to listen on [127.0.0.1]',
    )
    parser.add_argument(
        '-p',
        '--port',
        metavar='PORT',
        dest='port',
        type=int,
        default=5000,
        help='port to listen on [5000]',
    )
    parser.add_argument(
        '-Q',
        '--quality',
        metavar='QUALITY',
        dest='DEEPZOOM_TILE_QUALITY',
        type=int,
        help='JPEG compression quality [75]',
    )
    parser.add_argument(
        '-s',
        '--size',
        metavar='PIXELS',
        dest='DEEPZOOM_TILE_SIZE',
        type=int,
        help='tile size [254]',
    )
    parser.add_argument(
        'DEEPZOOM_SLIDE',
        metavar='SLIDE',
        type=Path,
        nargs='?',
        help='slide file',
    )

    args = parser.parse_args()
    config = {}
    config_file = args.config
    # Set only those settings specified on the command line
    for k in dir(args):
        v = getattr(args, k)
        if not k.startswith('_') and v is not None:
            config[k] = v
    app = create_app(config, config_file)

    app.run(host=args.host, port=args.port, threaded=True)



#
# openslide-python - Python bindings for the OpenSlide library
#
# Copyright (c) 2014 Carnegie Mellon University
# Copyright (c) 2024 Benjamin Gilbert
#
# This library is free software; you can redistribute it and/or modify it
# under the terms of version 2.1 of the GNU Lesser General Public License
# as published by the Free Software Foundation.
#
# This library is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public
# License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this library.  If not, see <https://www.gnu.org/licenses/>.
#

# Sphinx hardcodes that certain output paths have names starting with
# an underscore.
# Jekyll hardcodes that filenames starting with an underscore are not
# deployed to the website.
# Rename Sphinx output paths to drop the underscore.

from __future__ import annotations

import os
from pathlib import Path

from sphinx.application import Sphinx
from sphinx.util import logging
from sphinx.util.console import bold

DIRS = {
    '_static': 'static',
    '_sources': 'sources',
}
FILES = {
    # Added in Sphinx 5.0.0, scheduled to be removed in Sphinx 6
    'static/_sphinx_javascript_frameworks_compat.js': 'static/sphinx_javascript_frameworks_compat.js',  # noqa: E501
}
REWRITE_EXTENSIONS = {'.html', '.js'}


def remove_path_underscores(app: Sphinx, exception: Exception | None) -> None:
    if exception:
        return
    # Get logger
    logger = logging.getLogger(__name__)
    logger.info(bold('fixing pathnames... '), nonl=True)
    # Rewrite references in HTML/JS files
    outdir = Path(app.outdir)
    for dirpath, _, filenames in os.walk(outdir):
        for filename in filenames:
            path = Path(dirpath) / filename
            if path.suffix in REWRITE_EXTENSIONS:
                with path.open(encoding='utf-8') as fh:
                    contents = fh.read()
                for old, new in DIRS.items():
                    contents = contents.replace(old + '/', new + '/')
                for old, new in FILES.items():
                    contents = contents.replace(old, new)
                with path.open('w', encoding='utf-8') as fh:
                    fh.write(contents)
    # Move directory contents
    for old, new in DIRS.items():
        olddir = outdir / old
        newdir = outdir / new
        newdir.mkdir(exist_ok=True)
        if olddir.is_dir():
            for oldfile in olddir.iterdir():
                oldfile.rename(newdir / oldfile.name)
            olddir.rmdir()
    # Move files
    for old, new in FILES.items():
        oldfile = outdir / old
        if oldfile.is_file():
            oldfile.rename(outdir / new)
    logger.info('done')


def setup(app: Sphinx) -> None:
    app.connect('build-finished', remove_path_underscores)



#
# openslide-python - Python bindings for the OpenSlide library
#
# Copyright (c) 2016-2024 Benjamin Gilbert
#
# This library is free software; you can redistribute it and/or modify it
# under the terms of version 2.1 of the GNU Lesser General Public License
# as published by the Free Software Foundation.
#
# This library is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public
# License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this library.  If not, see <https://www.gnu.org/licenses/>.
#

from __future__ import annotations

import sys
import unittest

from PIL import Image
from common import file_path

from openslide import ImageSlide, OpenSlideCache, OpenSlideError, lowlevel


class TestImageWithoutOpening(unittest.TestCase):
    def test_detect_format(self) -> None:
        self.assertTrue(ImageSlide.detect_format(file_path('__missing_file')) is None)
        self.assertTrue(ImageSlide.detect_format(file_path('../setup.py')) is None)
        self.assertEqual(ImageSlide.detect_format(file_path('boxes.png')), 'PNG')

    def test_open(self) -> None:
        self.assertRaises(OSError, lambda: ImageSlide(file_path('__does_not_exist')))
        self.assertRaises(OSError, lambda: ImageSlide(file_path('../setup.py')))

    def test_open_image(self) -> None:
        # passing PIL.Image to ImageSlide
        with Image.open(file_path('boxes.png')) as img:
            with ImageSlide(img) as osr:
                self.assertEqual(osr.dimensions, (300, 250))
                self.assertEqual(repr(osr), 'ImageSlide(%r)' % img)

    @unittest.skipUnless(
        sys.getfilesystemencoding() == 'utf-8',
        'Python filesystem encoding is not UTF-8',
    )
    def test_unicode_path(self) -> None:
        path = file_path('😐.png')
        for arg in path, str(path):
            self.assertEqual(ImageSlide.detect_format(arg), 'PNG')
            self.assertEqual(ImageSlide(arg).dimensions, (300, 250))

    def test_unicode_path_bytes(self) -> None:
        arg = str(file_path('😐.png')).encode('UTF-8')
        self.assertEqual(ImageSlide.detect_format(arg), 'PNG')
        self.assertEqual(ImageSlide(arg).dimensions, (300, 250))

    def test_operations_on_closed_handle(self) -> None:
        with Image.open(file_path('boxes.png')) as img:
            osr = ImageSlide(img)
            osr.close()
            self.assertRaises(
                ValueError, lambda: osr.read_region((0, 0), 0, (100, 100))
            )
            self.assertRaises(ValueError, lambda: osr.level_dimensions)
            # If an Image is passed to the constructor, ImageSlide.close()
            # shouldn't close it
            self.assertEqual(img.getpixel((0, 0)), 3)

    def test_context_manager(self) -> None:
        osr = ImageSlide(file_path('boxes.png'))
        with osr:
            pass
        self.assertRaises(ValueError, lambda: osr.read_region((0, 0), 0, (100, 100)))
        self.assertRaises(ValueError, lambda: osr.level_dimensions)


class _Abstract:
    # nested class to prevent the test runner from finding it
    class SlideTest(unittest.TestCase):
        FILENAME: str | None = None

        def setUp(self) -> None:
            assert self.FILENAME is not None
            self.osr = ImageSlide(file_path(self.FILENAME))

        def tearDown(self) -> None:
            self.osr.close()


class TestImage(_Abstract.SlideTest):
    FILENAME = 'boxes.png'

    def test_repr(self) -> None:
        self.assertEqual(repr(self.osr), 'ImageSlide(%r)' % file_path('boxes.png'))

    def test_metadata(self) -> None:
        self.assertEqual(self.osr.level_count, 1)
        self.assertEqual(self.osr.level_dimensions, ((300, 250),))
        self.assertEqual(self.osr.dimensions, (300, 250))
        self.assertEqual(self.osr.level_downsamples, (1.0,))

        self.assertEqual(self.osr.get_best_level_for_downsample(0.5), 0)
        self.assertEqual(self.osr.get_best_level_for_downsample(3), 0)

        self.assertEqual(self.osr.properties, {})
        self.assertEqual(self.osr.associated_images, {})

    def test_color_profile(self) -> None:
        assert self.osr.color_profile is not None  # for type inference
        self.assertEqual(self.osr.color_profile.profile.device_class, 'mntr')
        self.assertEqual(
            len(self.osr.read_region((0, 0), 0, (100, 100)).info['icc_profile']), 588
        )
        self.assertEqual(
            len(self.osr.get_thumbnail((100, 100)).info['icc_profile']), 588
        )

    def test_read_region(self) -> None:
        self.assertEqual(
            self.osr.read_region((-10, -10), 0, (400, 400)).size, (400, 400)
        )

    def test_read_region_size_dimension_zero(self) -> None:
        self.assertEqual(self.osr.read_region((0, 0), 0, (400, 0)).size, (400, 0))

    def test_read_region_bad_level(self) -> None:
        self.assertRaises(
            OpenSlideError, lambda: self.osr.read_region((0, 0), 1, (100, 100))
        )

    def test_read_region_bad_size(self) -> None:
        self.assertRaises(
            OpenSlideError, lambda: self.osr.read_region((0, 0), 0, (400, -5))
        )

    def test_thumbnail(self) -> None:
        self.assertEqual(self.osr.get_thumbnail((100, 100)).size, (100, 83))

    @unittest.skipUnless(lowlevel.cache_create.available, "requires OpenSlide 4.0.0")
    def test_set_cache(self) -> None:
        self.osr.set_cache(OpenSlideCache(64 << 10))
        self.assertEqual(self.osr.read_region((0, 0), 0, (400, 400)).size, (400, 400))


class TestNoIccImage(_Abstract.SlideTest):
    FILENAME = 'boxes-no-icc.png'

    def test_color_profile(self) -> None:
        self.assertIsNone(self.osr.color_profile)
        self.assertNotIn(
            'icc_profile', self.osr.read_region((0, 0), 0, (100, 100)).info
        )
        self.assertNotIn('icc_profile', self.osr.get_thumbnail((100, 100)).info)



#
# openslide-python - Python bindings for the OpenSlide library
#
# Copyright (c) 2010-2014 Carnegie Mellon University
#
# This library is free software; you can redistribute it and/or modify it
# under the terms of version 2.1 of the GNU Lesser General Public License
# as published by the Free Software Foundation.
#
# This library is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public
# License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this library.  If not, see <https://www.gnu.org/licenses/>.
#

"""Support for Deep Zoom images.

This module provides functionality for generating Deep Zoom images from
OpenSlide objects.
"""

from __future__ import annotations

from io import BytesIO
import math
from typing import TYPE_CHECKING
from xml.etree.ElementTree import Element, ElementTree, SubElement

from PIL import Image

import openslide

if TYPE_CHECKING:
    # Python 3.10+
    from typing import TypeGuard


class DeepZoomGenerator:
    """Generates Deep Zoom tiles and metadata."""

    BOUNDS_OFFSET_PROPS = (
        openslide.PROPERTY_NAME_BOUNDS_X,
        openslide.PROPERTY_NAME_BOUNDS_Y,
    )
    BOUNDS_SIZE_PROPS = (
        openslide.PROPERTY_NAME_BOUNDS_WIDTH,
        openslide.PROPERTY_NAME_BOUNDS_HEIGHT,
    )

    def __init__(
        self,
        osr: openslide.AbstractSlide,
        tile_size: int = 254,
        overlap: int = 1,
        limit_bounds: bool = False,
    ):
        """Create a DeepZoomGenerator wrapping an OpenSlide object.

        osr:          a slide object.
        tile_size:    the width and height of a single tile.  For best viewer
                      performance, tile_size + 2 * overlap should be a power
                      of two.
        overlap:      the number of extra pixels to add to each interior edge
                      of a tile.
        limit_bounds: True to render only the non-empty slide region."""

        # We have four coordinate planes:
        # - Row and column of the tile within the Deep Zoom level (t_)
        # - Pixel coordinates within the Deep Zoom level (z_)
        # - Pixel coordinates within the slide level (l_)
        # - Pixel coordinates within slide level 0 (l0_)

        self._osr = osr
        self._z_t_downsample = tile_size
        self._z_overlap = overlap
        self._limit_bounds = limit_bounds

        # Precompute dimensions
        # Slide level and offset
        if limit_bounds:
            # Level 0 coordinate offset
            self._l0_offset = tuple(
                int(osr.properties.get(prop, 0)) for prop in self.BOUNDS_OFFSET_PROPS
            )
            # Slide level dimensions scale factor in each axis
            size_scale = tuple(
                int(osr.properties.get(prop, l0_lim)) / l0_lim
                for prop, l0_lim in zip(self.BOUNDS_SIZE_PROPS, osr.dimensions)
            )
            # Dimensions of active area
            self._l_dimensions = tuple(
                tuple(
                    int(math.ceil(l_lim * scale))
                    for l_lim, scale in zip(l_size, size_scale)
                )
                for l_size in osr.level_dimensions
            )
        else:
            self._l_dimensions = osr.level_dimensions
            self._l0_offset = (0, 0)
        self._l0_dimensions = self._l_dimensions[0]
        # Deep Zoom level
        z_size = self._l0_dimensions
        z_dimensions = [z_size]
        while z_size[0] > 1 or z_size[1] > 1:
            z_size = tuple(max(1, int(math.ceil(z / 2))) for z in z_size)
            z_dimensions.append(z_size)
        # Narrow the type, for self.level_dimensions
        self._z_dimensions = self._pairs_from_n_tuples(tuple(reversed(z_dimensions)))

        # Tile
        def tiles(z_lim: int) -> int:
            return int(math.ceil(z_lim / self._z_t_downsample))

        self._t_dimensions = tuple(
            (tiles(z_w), tiles(z_h)) for z_w, z_h in self._z_dimensions
        )

        # Deep Zoom level count
        self._dz_levels = len(self._z_dimensions)

        # Total downsamples for each Deep Zoom level
        # mypy infers this as a tuple[Any, ...] due to the ** operator
        l0_z_downsamples: tuple[int, ...] = tuple(
            2 ** (self._dz_levels - dz_level - 1) for dz_level in range(self._dz_levels)
        )

        # Preferred slide levels for each Deep Zoom level
        self._slide_from_dz_level = tuple(
            self._osr.get_best_level_for_downsample(d) for d in l0_z_downsamples
        )

        # Piecewise downsamples
        self._l0_l_downsamples = self._osr.level_downsamples
        self._l_z_downsamples = tuple(
            l0_z_downsamples[dz_level]
            / self._l0_l_downsamples[self._slide_from_dz_level[dz_level]]
            for dz_level in range(self._dz_levels)
        )

        # Slide background color
        self._bg_color = '#' + self._osr.properties.get(
            openslide.PROPERTY_NAME_BACKGROUND_COLOR, 'ffffff'
        )

    def __repr__(self) -> str:
        return '{}({!r}, tile_size={!r}, overlap={!r}, limit_bounds={!r})'.format(
            self.__class__.__name__,
            self._osr,
            self._z_t_downsample,
            self._z_overlap,
            self._limit_bounds,
        )

    @property
    def level_count(self) -> int:
        """The number of Deep Zoom levels in the image."""
        return self._dz_levels

    @property
    def level_tiles(self) -> tuple[tuple[int, int], ...]:
        """A tuple of (tiles_x, tiles_y) tuples for each Deep Zoom level."""
        return self._t_dimensions

    @property
    def level_dimensions(self) -> tuple[tuple[int, int], ...]:
        """A tuple of (pixels_x, pixels_y) tuples for each Deep Zoom level."""
        return self._z_dimensions

    @property
    def tile_count(self) -> int:
        """The total number of Deep Zoom tiles in the image."""
        return sum(t_cols * t_rows for t_cols, t_rows in self._t_dimensions)

    def get_tile(self, level: int, address: tuple[int, int]) -> Image.Image:
        """Return an RGB PIL.Image for a tile.

        level:     the Deep Zoom level.
        address:   the address of the tile within the level as a (col, row)
                   tuple."""

        # Read tile
        args, z_size = self._get_tile_info(level, address)
        tile = self._osr.read_region(*args)
        profile = tile.info.get('icc_profile')

        # Apply on solid background
        bg = Image.new('RGB', tile.size, self._bg_color)
        tile = Image.composite(tile, bg, tile)

        # Scale to the correct size
        if tile.size != z_size:
            # Image.Resampling added in Pillow 9.1.0
            # Image.LANCZOS removed in Pillow 10
            tile.thumbnail(z_size, getattr(Image, 'Resampling', Image).LANCZOS)

        # Reference ICC profile
        if profile is not None:
            tile.info['icc_profile'] = profile

        return tile

    def _get_tile_info(
        self, dz_level: int, t_location: tuple[int, int]
    ) -> tuple[tuple[tuple[int, int], int, tuple[int, int]], tuple[int, int]]:
        # Check parameters
        if dz_level < 0 or dz_level >= self._dz_levels:
            raise ValueError("Invalid level")
        for t, t_lim in zip(t_location, self._t_dimensions[dz_level]):
            if t < 0 or t >= t_lim:
                raise ValueError("Invalid address")

        # Get preferred slide level
        slide_level = self._slide_from_dz_level[dz_level]

        # Calculate top/left and bottom/right overlap
        z_overlap_tl = tuple(self._z_overlap * int(t != 0) for t in t_location)
        z_overlap_br = tuple(
            self._z_overlap * int(t != t_lim - 1)
            for t, t_lim in zip(t_location, self.level_tiles[dz_level])
        )

        # Get final size of the tile
        z_size = tuple(
            min(self._z_t_downsample, z_lim - self._z_t_downsample * t) + z_tl + z_br
            for t, z_lim, z_tl, z_br in zip(
                t_location, self._z_dimensions[dz_level], z_overlap_tl, z_overlap_br
            )
        )

        # Obtain the region coordinates
        z_location = [self._z_from_t(t) for t in t_location]
        l_location = [
            self._l_from_z(dz_level, z - z_tl)
            for z, z_tl in zip(z_location, z_overlap_tl)
        ]
        # Round location down and size up, and add offset of active area
        l0_location = tuple(
            int(self._l0_from_l(slide_level, l) + l0_off)
            for l, l0_off in zip(l_location, self._l0_offset)
        )
        l_size = tuple(
            int(min(math.ceil(self._l_from_z(dz_level, dz)), l_lim - math.ceil(l)))
            for l, dz, l_lim in zip(l_location, z_size, self._l_dimensions[slide_level])
        )

        # Return read_region() parameters plus tile size for final scaling
        assert len(l0_location) == 2 and len(l_size) == 2 and len(z_size) == 2
        return ((l0_location, slide_level, l_size), z_size)

    def _l0_from_l(self, slide_level: int, l: float) -> float:
        return self._l0_l_downsamples[slide_level] * l

    def _l_from_z(self, dz_level: int, z: int) -> float:
        return self._l_z_downsamples[dz_level] * z

    def _z_from_t(self, t: int) -> int:
        return self._z_t_downsample * t

    @staticmethod
    def _pairs_from_n_tuples(
        tuples: tuple[tuple[int, ...], ...],
    ) -> tuple[tuple[int, int], ...]:
        def all_pairs(
            tuples: tuple[tuple[int, ...], ...],
        ) -> TypeGuard[tuple[tuple[int, int], ...]]:
            return all(len(t) == 2 for t in tuples)

        assert all_pairs(tuples)
        return tuples

    def get_tile_coordinates(
        self, level: int, address: tuple[int, int]
    ) -> tuple[tuple[int, int], int, tuple[int, int]]:
        """Return the OpenSlide.read_region() arguments for the specified tile.

        Most users should call get_tile() rather than calling
        OpenSlide.read_region() directly.

        level:     the Deep Zoom level.
        address:   the address of the tile within the level as a (col, row)
                   tuple."""
        return self._get_tile_info(level, address)[0]

    def get_tile_dimensions(
        self, level: int, address: tuple[int, int]
    ) -> tuple[int, int]:
        """Return a (pixels_x, pixels_y) tuple for the specified tile.

        level:     the Deep Zoom level.
        address:   the address of the tile within the level as a (col, row)
                   tuple."""
        return self._get_tile_info(level, address)[1]

    def get_dzi(self, format: str) -> str:
        """Return a string containing the XML metadata for the .dzi file.

        format:    the format of the individual tiles ('png' or 'jpeg')"""
        image = Element(
            'Image',
            TileSize=str(self._z_t_downsample),
            Overlap=str(self._z_overlap),
            Format=format,
            xmlns='http://schemas.microsoft.com/deepzoom/2008',
        )
        w, h = self._l0_dimensions
        SubElement(image, 'Size', Width=str(w), Height=str(h))
        tree = ElementTree(element=image)
        buf = BytesIO()
        tree.write(buf, encoding='UTF-8')
        return buf.getvalue().decode('UTF-8')



#
# openslide-python - Python bindings for the OpenSlide library
#
# Copyright (c) 2010-2014 Carnegie Mellon University
# Copyright (c) 2021-2023 Benjamin Gilbert
#
# This library is free software; you can redistribute it and/or modify it
# under the terms of version 2.1 of the GNU Lesser General Public License
# as published by the Free Software Foundation.
#
# This library is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public
# License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this library.  If not, see <https://www.gnu.org/licenses/>.
#

"""A library for reading whole-slide images.

This package provides Python bindings for the OpenSlide library.
"""

from __future__ import annotations

from abc import ABCMeta, abstractmethod
from collections.abc import Iterator, Mapping
from io import BytesIO
from types import TracebackType
from typing import Literal, TypeVar

from PIL import Image, ImageCms

from openslide import lowlevel

# Re-exports for the benefit of library users
from openslide._version import (  # noqa: F401  module-imported-but-unused
    __version__ as __version__,
)
from openslide.lowlevel import (
    OpenSlideUnsupportedFormatError as OpenSlideUnsupportedFormatError,
)
from openslide.lowlevel import (  # noqa: F401  module-imported-but-unused
    OpenSlideVersionError as OpenSlideVersionError,
)
from openslide.lowlevel import OpenSlideError as OpenSlideError

__library_version__ = lowlevel.get_version()

PROPERTY_NAME_COMMENT = 'openslide.comment'
PROPERTY_NAME_VENDOR = 'openslide.vendor'
PROPERTY_NAME_QUICKHASH1 = 'openslide.quickhash-1'
PROPERTY_NAME_BACKGROUND_COLOR = 'openslide.background-color'
PROPERTY_NAME_OBJECTIVE_POWER = 'openslide.objective-power'
PROPERTY_NAME_MPP_X = 'openslide.mpp-x'
PROPERTY_NAME_MPP_Y = 'openslide.mpp-y'
PROPERTY_NAME_BOUNDS_X = 'openslide.bounds-x'
PROPERTY_NAME_BOUNDS_Y = 'openslide.bounds-y'
PROPERTY_NAME_BOUNDS_WIDTH = 'openslide.bounds-width'
PROPERTY_NAME_BOUNDS_HEIGHT = 'openslide.bounds-height'

_T = TypeVar('_T')


class AbstractSlide(metaclass=ABCMeta):
    """The base class of a slide object."""

    def __init__(self) -> None:
        self._profile: bytes | None = None

    def __enter__(self: _T) -> _T:
        return self

    def __exit__(
        self,
        exc_type: type[BaseException] | None,
        exc_value: BaseException | None,
        traceback: TracebackType | None,
    ) -> Literal[False]:
        self.close()
        return False

    @classmethod
    @abstractmethod
    def detect_format(cls, filename: lowlevel.Filename) -> str | None:
        """Return a string describing the format of the specified file.

        If the file format is not recognized, return None."""
        raise NotImplementedError

    @abstractmethod
    def close(self) -> None:
        """Close the slide."""
        raise NotImplementedError

    @property
    @abstractmethod
    def level_count(self) -> int:
        """The number of levels in the image."""
        raise NotImplementedError

    @property
    @abstractmethod
    def level_dimensions(self) -> tuple[tuple[int, int], ...]:
        """A tuple of (width, height) tuples, one for each level of the image.

        level_dimensions[n] contains the dimensions of level n."""
        raise NotImplementedError

    @property
    def dimensions(self) -> tuple[int, int]:
        """A (width, height) tuple for level 0 of the image."""
        return self.level_dimensions[0]

    @property
    @abstractmethod
    def level_downsamples(self) -> tuple[float, ...]:
        """A tuple of downsampling factors for each level of the image.

        level_downsample[n] contains the downsample factor of level n."""
        raise NotImplementedError

    @property
    @abstractmethod
    def properties(self) -> Mapping[str, str]:
        """Metadata about the image.

        This is a map: property name -> property value."""
        raise NotImplementedError

    @property
    @abstractmethod
    def associated_images(self) -> Mapping[str, Image.Image]:
        """Images associated with this whole-slide image.

        This is a map: image name -> PIL.Image."""
        raise NotImplementedError

    @property
    def color_profile(self) -> ImageCms.ImageCmsProfile | None:
        """Color profile for the whole-slide image, or None if unavailable."""
        if self._profile is None:
            return None
        return ImageCms.getOpenProfile(BytesIO(self._profile))

    @abstractmethod
    def get_best_level_for_downsample(self, downsample: float) -> int:
        """Return the best level for displaying the given downsample."""
        raise NotImplementedError

    @abstractmethod
    def read_region(
        self, location: tuple[int, int], level: int, size: tuple[int, int]
    ) -> Image.Image:
        """Return a PIL.Image containing the contents of the region.

        location: (x, y) tuple giving the top left pixel in the level 0
                  reference frame.
        level:    the level number.
        size:     (width, height) tuple giving the region size."""
        raise NotImplementedError

    def set_cache(self, cache: OpenSlideCache) -> None:  # noqa: B027
        """Use the specified cache to store recently decoded slide tiles.

        This class does not support caching, so this method does nothing.

        cache: an OpenSlideCache object."""
        pass

    def get_thumbnail(self, size: tuple[int, int]) -> Image.Image:
        """Return a PIL.Image containing an RGB thumbnail of the image.

        size:     the maximum size of the thumbnail."""
        downsample = max(dim / thumb for dim, thumb in zip(self.dimensions, size))
        level = self.get_best_level_for_downsample(downsample)
        tile = self.read_region((0, 0), level, self.level_dimensions[level])
        # Apply on solid background
        bg_color = '#' + self.properties.get(PROPERTY_NAME_BACKGROUND_COLOR, 'ffffff')
        thumb = Image.new('RGB', tile.size, bg_color)
        thumb.paste(tile, None, tile)
        # Image.Resampling added in Pillow 9.1.0
        # Image.LANCZOS removed in Pillow 10
        thumb.thumbnail(size, getattr(Image, 'Resampling', Image).LANCZOS)
        if self._profile is not None:
            thumb.info['icc_profile'] = self._profile
        return thumb


class OpenSlide(AbstractSlide):
    """An open whole-slide image.

    close() is called automatically when the object is deleted.
    The object may be used as a context manager, in which case it will be
    closed upon exiting the context.

    If an operation fails, OpenSlideError is raised.  Note that OpenSlide
    has latching error semantics: once OpenSlideError is raised, all future
    operations on the OpenSlide object, other than close(), will fail.
    """

    def __init__(self, filename: lowlevel.Filename):
        """Open a whole-slide image."""
        AbstractSlide.__init__(self)
        self._filename = filename
        self._osr = lowlevel.open(filename)
        if lowlevel.read_icc_profile.available:
            self._profile = lowlevel.read_icc_profile(self._osr)

    def __repr__(self) -> str:
        return f'{self.__class__.__name__}({self._filename!r})'

    @classmethod
    def detect_format(cls, filename: lowlevel.Filename) -> str | None:
        """Return a string describing the format vendor of the specified file.

        If the file format is not recognized, return None."""
        return lowlevel.detect_vendor(filename)

    def close(self) -> None:
        """Close the OpenSlide object."""
        lowlevel.close(self._osr)

    @property
    def level_count(self) -> int:
        """The number of levels in the image."""
        return lowlevel.get_level_count(self._osr)

    @property
    def level_dimensions(self) -> tuple[tuple[int, int], ...]:
        """A tuple of (width, height) tuples, one for each level of the image.

        level_dimensions[n] contains the dimensions of level n."""
        return tuple(
            lowlevel.get_level_dimensions(self._osr, i) for i in range(self.level_count)
        )

    @property
    def level_downsamples(self) -> tuple[float, ...]:
        """A tuple of downsampling factors for each level of the image.

        level_downsample[n] contains the downsample factor of level n."""
        return tuple(
            lowlevel.get_level_downsample(self._osr, i) for i in range(self.level_count)
        )

    @property
    def properties(self) -> Mapping[str, str]:
        """Metadata about the image.

        This is a map: property name -> property value."""
        return _PropertyMap(self._osr)

    @property
    def associated_images(self) -> Mapping[str, Image.Image]:
        """Images associated with this whole-slide image.

        This is a map: image name -> PIL.Image.

        Unlike in the C interface, the images accessible via this property
        are not premultiplied."""
        return _AssociatedImageMap(self._osr, self._profile)

    def get_best_level_for_downsample(self, downsample: float) -> int:
        """Return the best level for displaying the given downsample."""
        return lowlevel.get_best_level_for_downsample(self._osr, downsample)

    def read_region(
        self, location: tuple[int, int], level: int, size: tuple[int, int]
    ) -> Image.Image:
        """Return a PIL.Image containing the contents of the region.

        location: (x, y) tuple giving the top left pixel in the level 0
                  reference frame.
        level:    the level number.
        size:     (width, height) tuple giving the region size.

        Unlike in the C interface, the image data returned by this
        function is not premultiplied."""
        region = lowlevel.read_region(
            self._osr, location[0], location[1], level, size[0], size[1]
        )
        if self._profile is not None:
            region.info['icc_profile'] = self._profile
        return region

    def set_cache(self, cache: OpenSlideCache) -> None:
        """Use the specified cache to store recently decoded slide tiles.

        By default, the object has a private cache with a default size.

        cache: an OpenSlideCache object."""
        try:
            llcache = cache._openslide_cache
        except AttributeError as exc:
            raise TypeError('Not a cache object') from exc
        lowlevel.set_cache(self._osr, llcache)


class _OpenSlideMap(Mapping[str, _T]):
    def __init__(self, osr: lowlevel._OpenSlide):
        self._osr = osr

    def __repr__(self) -> str:
        return f'<{self.__class__.__name__} {dict(self)!r}>'

    def __len__(self) -> int:
        return len(self._keys())

    def __iter__(self) -> Iterator[str]:
        return iter(self._keys())

    @abstractmethod
    def _keys(self) -> list[str]:
        # Private method; always returns list.
        raise NotImplementedError()


class _PropertyMap(_OpenSlideMap[str]):
    def _keys(self) -> list[str]:
        return lowlevel.get_property_names(self._osr)

    def __getitem__(self, key: str) -> str:
        v = lowlevel.get_property_value(self._osr, key)
        if v is None:
            raise KeyError()
        return v


class _AssociatedImageMap(_OpenSlideMap[Image.Image]):
    def __init__(self, osr: lowlevel._OpenSlide, profile: bytes | None):
        _OpenSlideMap.__init__(self, osr)
        self._profile = profile

    def _keys(self) -> list[str]:
        return lowlevel.get_associated_image_names(self._osr)

    def __getitem__(self, key: str) -> Image.Image:
        if key not in self._keys():
            raise KeyError()
        image = lowlevel.read_associated_image(self._osr, key)
        if lowlevel.read_associated_image_icc_profile.available:
            profile = lowlevel.read_associated_image_icc_profile(self._osr, key)
            if profile == self._profile:
                # reuse profile copy from main image to save memory
                profile = self._profile
            if profile is not None:
                image.info['icc_profile'] = profile
        return image


class OpenSlideCache:
    """An in-memory tile cache.

    Tile caches can be attached to one or more OpenSlide objects with
    OpenSlide.set_cache() to cache recently-decoded tiles.  By default,
    each OpenSlide object has its own cache with a default size.
    """

    def __init__(self, capacity: int):
        """Create a tile cache with the specified capacity in bytes."""
        self._capacity = capacity
        self._openslide_cache = lowlevel.cache_create(capacity)

    def __repr__(self) -> str:
        return f'{self.__class__.__name__}({self._capacity!r})'


class ImageSlide(AbstractSlide):
    """A wrapper for a PIL.Image that provides the OpenSlide interface."""

    def __init__(self, file: lowlevel.Filename | Image.Image):
        """Open an image file.

        file can be a filename or a PIL.Image."""
        AbstractSlide.__init__(self)
        self._file_arg = file
        if isinstance(file, Image.Image):
            self._close = False
            self._image: Image.Image | None = file
        else:
            self._close = True
            self._image = Image.open(file)
        self._profile = self._image.info.get('icc_profile')

    def __repr__(self) -> str:
        return f'{self.__class__.__name__}({self._file_arg!r})'

    @classmethod
    def detect_format(cls, filename: lowlevel.Filename) -> str | None:
        """Return a string describing the format of the specified file.

        If the file format is not recognized, return None."""
        try:
            with Image.open(filename) as img:
                # img currently resolves as Any
                # https://github.com/python-pillow/Pillow/pull/8362
                return img.format  # type: ignore[no-any-return]
        except OSError:
            return None

    def close(self) -> None:
        """Close the slide object."""
        if self._close:
            assert self._image is not None
            self._image.close()
            self._close = False
        self._image = None

    @property
    def level_count(self) -> Literal[1]:
        """The number of levels in the image."""
        return 1

    @property
    def level_dimensions(self) -> tuple[tuple[int, int]]:
        """A tuple of (width, height) tuples, one for each level of the image.

        level_dimensions[n] contains the dimensions of level n."""
        if self._image is None:
            raise ValueError('Cannot read from a closed slide')
        return (self._image.size,)

    @property
    def level_downsamples(self) -> tuple[float]:
        """A tuple of downsampling factors for each level of the image.

        level_downsample[n] contains the downsample factor of level n."""
        return (1.0,)

    @property
    def properties(self) -> Mapping[str, str]:
        """Metadata about the image.

        This is a map: property name -> property value."""
        return {}

    @property
    def associated_images(self) -> Mapping[str, Image.Image]:
        """Images associated with this whole-slide image.

        This is a map: image name -> PIL.Image."""
        return {}

    def get_best_level_for_downsample(self, _downsample: float) -> Literal[0]:
        """Return the best level for displaying the given downsample."""
        return 0

    def read_region(
        self, location: tuple[int, int], level: int, size: tuple[int, int]
    ) -> Image.Image:
        """Return a PIL.Image containing the contents of the region.

        location: (x, y) tuple giving the top left pixel in the level 0
                  reference frame.
        level:    the level number.
        size:     (width, height) tuple giving the region size."""
        if self._image is None:
            raise ValueError('Cannot read from a closed slide')
        if level != 0:
            raise OpenSlideError("Invalid level")
        if ['fail' for s in size if s < 0]:
            raise OpenSlideError(f"Size {size} must be non-negative")
        # Any corner of the requested region may be outside the bounds of
        # the image.  Create a transparent tile of the correct size and
        # paste the valid part of the region into the correct location.
        image_topleft = [
            max(0, min(l, limit - 1)) for l, limit in zip(location, self._image.size)
        ]
        image_bottomright = [
            max(0, min(l + s - 1, limit - 1))
            for l, s, limit in zip(location, size, self._image.size)
        ]
        tile = Image.new("RGBA", size, (0,) * 4)
        if not [
            'fail' for tl, br in zip(image_topleft, image_bottomright) if br - tl < 0
        ]:  # "< 0" not a typo
            # Crop size is greater than zero in both dimensions.
            # PIL thinks the bottom right is the first *excluded* pixel
            crop_box = tuple(image_topleft + [d + 1 for d in image_bottomright])
            tile_offset = tuple(il - l for il, l in zip(image_topleft, location))
            assert len(crop_box) == 4 and len(tile_offset) == 2
            crop = self._image.crop(crop_box)
            tile.paste(crop, tile_offset)
        if self._profile is not None:
            tile.info['icc_profile'] = self._profile
        return tile


def open_slide(filename: lowlevel.Filename) -> OpenSlide | ImageSlide:
    """Open a whole-slide or regular image.

    Return an OpenSlide object for whole-slide images and an ImageSlide
    object for other types of images."""
    try:
        return OpenSlide(filename)
    except OpenSlideUnsupportedFormatError:
        return ImageSlide(filename)


if __name__ == '__main__':
    import sys

    print("OpenSlide vendor:", OpenSlide.detect_format(sys.argv[1]))
    print("PIL format:", ImageSlide.detect_format(sys.argv[1]))
    with open_slide(sys.argv[1]) as _slide:
        print("Dimensions:", _slide.dimensions)
        print("Levels:", _slide.level_count)
        print("Level dimensions:", _slide.level_dimensions)
        print("Level downsamples:", _slide.level_downsamples)
        print("Properties:", _slide.properties)
        print("Associated images:", _slide.associated_images)



#
# openslide-python - Python bindings for the OpenSlide library
#
# Copyright (c) 2016-2021 Benjamin Gilbert
#
# This library is free software; you can redistribute it and/or modify it
# under the terms of version 2.1 of the GNU Lesser General Public License
# as published by the Free Software Foundation.
#
# This library is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public
# License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this library.  If not, see <https://www.gnu.org/licenses/>.
#

from __future__ import annotations

import os
from pathlib import Path

# Handle Windows-specific first-import logic here, so individual modules
# don't have to
if os.name == 'nt':
    # In application code, you probably shouldn't use an environment
    # variable for this, unless you're sure you can trust the contents of the
    # environment.
    _dll_path = os.getenv('OPENSLIDE_PATH')
    if _dll_path is not None:
        with os.add_dll_directory(_dll_path):  # type: ignore[attr-defined,unused-ignore]  # noqa: E501
            import openslide  # noqa: F401  module-imported-but-unused


def file_path(name: str) -> Path:
    return Path(__file__).parent / 'fixtures' / name



#
# openslide-python - Python bindings for the OpenSlide library
#
# Copyright (c) 2010-2013 Carnegie Mellon University
# Copyright (c) 2016-2024 Benjamin Gilbert
#
# This library is free software; you can redistribute it and/or modify it
# under the terms of version 2.1 of the GNU Lesser General Public License
# as published by the Free Software Foundation.
#
# This library is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public
# License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this library.  If not, see <https://www.gnu.org/licenses/>.
#

"""
Low-level interface to the OpenSlide library.

Most users should use the openslide.OpenSlide class rather than this
module.

This module provides nearly direct equivalents to the OpenSlide C API.
(As an implementation detail, conversion of premultiplied image data
returned by OpenSlide into a non-premultiplied PIL.Image happens here
rather than in the high-level interface.)
"""

from __future__ import annotations

from collections.abc import Callable
from ctypes import (
    CDLL,
    POINTER,
    _Pointer,
    byref,
    c_char,
    c_char_p,
    c_double,
    c_int32,
    c_int64,
    c_size_t,
    c_uint32,
    c_void_p,
    cdll,
)
from itertools import count
import os
import platform
from typing import TYPE_CHECKING, Any, Protocol, TypeVar, cast

from PIL import Image

from . import _convert

if TYPE_CHECKING:
    # Python 3.10+
    from typing import ParamSpec, TypeAlias

    from _convert import _Buffer


def _load_library() -> CDLL:
    try:
        import openslide_bin

        return openslide_bin.libopenslide1
    except (AttributeError, ModuleNotFoundError):
        pass

    def try_load(names: list[str]) -> CDLL:
        try:
            return cdll.LoadLibrary(names[0])
        except OSError:
            remaining = names[1:]
            if remaining:
                # handle recursively so implicit exception chaining captures
                # all the failures
                return try_load(remaining)
            else:
                raise

    if platform.system() == 'Windows':
        try:
            return try_load(['libopenslide-1.dll', 'libopenslide-0.dll'])
        except FileNotFoundError as exc:
            raise ModuleNotFoundError(
                "Couldn't locate OpenSlide DLL. "
                "Try `pip install openslide-bin`, "
                "or if you're using an OpenSlide binary package, "
                "ensure you've called os.add_dll_directory(). "
                "https://openslide.org/api/python/#installing"
            ) from exc
    elif platform.system() == 'Darwin':
        try:
            return try_load(['libopenslide.1.dylib', 'libopenslide.0.dylib'])
        except OSError as exc:
            # MacPorts doesn't add itself to the dyld search path, but
            # does add itself to the find_library() search path
            # (DEFAULT_LIBRARY_FALLBACK in ctypes.macholib.dyld).
            import ctypes.util

            lib = ctypes.util.find_library('openslide')
            if lib is None:
                raise ModuleNotFoundError(
                    "Couldn't locate OpenSlide dylib. "
                    "Try `pip install openslide-bin`. "
                    "https://openslide.org/api/python/#installing"
                ) from exc
            return cdll.LoadLibrary(lib)
    else:
        try:
            return try_load(['libopenslide.so.1', 'libopenslide.so.0'])
        except OSError as exc:
            raise ModuleNotFoundError(
                "Couldn't locate OpenSlide shared library. "
                "Try `pip install openslide-bin`. "
                "https://openslide.org/api/python/#installing"
            ) from exc


_lib = _load_library()


class OpenSlideError(Exception):
    """An error produced by the OpenSlide library.

    Import this from openslide rather than from openslide.lowlevel.
    """


class OpenSlideVersionError(OpenSlideError):
    """This version of OpenSlide does not support the requested functionality.

    Import this from openslide rather than from openslide.lowlevel.
    """

    def __init__(self, minimum_version: str):
        super().__init__(f'OpenSlide >= {minimum_version} required')
        self.minimum_version = minimum_version


class OpenSlideUnsupportedFormatError(OpenSlideError):
    """OpenSlide does not support the requested file.

    Import this from openslide rather than from openslide.lowlevel.
    """


class _OpenSlide:
    """Wrapper class to make sure we correctly pass an OpenSlide handle."""

    def __init__(self, ptr: c_void_p):
        self._as_parameter_ = ptr
        self._valid = True
        # Retain a reference to close() to avoid GC problems during
        # interpreter shutdown
        self._close = close

    def __del__(self) -> None:
        if self._valid:
            self._close(self)

    def invalidate(self) -> None:
        self._valid = False

    @classmethod
    def from_param(cls, obj: _OpenSlide) -> _OpenSlide:
        if obj.__class__ != cls:
            raise ValueError("Not an OpenSlide reference")
        if not obj._as_parameter_:
            raise ValueError("Passing undefined slide object")
        if not obj._valid:
            raise ValueError("Passing closed slide object")
        return obj


class _OpenSlideCache:
    """Wrapper class to make sure we correctly pass an OpenSlide cache."""

    def __init__(self, ptr: c_void_p):
        self._as_parameter_ = ptr
        # Retain a reference to cache_release() to avoid GC problems during
        # interpreter shutdown
        self._cache_release = cache_release

    def __del__(self) -> None:
        self._cache_release(self)

    @classmethod
    def from_param(cls, obj: _OpenSlideCache) -> _OpenSlideCache:
        if obj.__class__ != cls:
            raise ValueError("Not an OpenSlide cache reference")
        if not obj._as_parameter_:
            raise ValueError("Passing undefined cache object")
        return obj


if TYPE_CHECKING:
    # Python 3.10+
    Filename: TypeAlias = str | bytes | os.PathLike[Any]


class _filename_p:
    """Wrapper class to convert filename arguments to bytes."""

    @classmethod
    def from_param(cls, obj: Filename) -> bytes:
        # fspath and fsencode raise TypeError on unexpected types
        if platform.system() == 'Windows':
            # OpenSlide 4.0.0+ requires UTF-8 on Windows
            obj = os.fspath(obj)
            if isinstance(obj, str):
                return obj.encode('UTF-8')
            else:
                return obj
        else:
            return os.fsencode(obj)


class _utf8_p:
    """Wrapper class to convert string arguments to bytes."""

    @classmethod
    def from_param(cls, obj: str | bytes) -> bytes:
        if isinstance(obj, bytes):
            return obj
        elif isinstance(obj, str):
            return obj.encode('UTF-8')
        else:
            raise TypeError('Incorrect type')


class _size_t:
    """Wrapper class to convert size_t arguments to c_size_t."""

    @classmethod
    def from_param(cls, obj: int) -> c_size_t:
        if not isinstance(obj, int):
            raise TypeError('Incorrect type')
        if obj < 0:
            raise ValueError('Value out of range')
        return c_size_t(obj)


def _load_image(buf: _Buffer, size: tuple[int, int]) -> Image.Image:
    '''buf must be a mutable buffer.'''
    _convert.argb2rgba(buf)
    return Image.frombuffer('RGBA', size, buf, 'raw', 'RGBA', 0, 1)


# check for errors opening an image file and wrap the resulting handle
def _check_open(result: int | None, _func: Any, _args: Any) -> _OpenSlide:
    if result is None:
        raise OpenSlideUnsupportedFormatError("Unsupported or missing image file")
    slide = _OpenSlide(c_void_p(result))
    err = get_error(slide)
    if err is not None:
        raise OpenSlideError(err)
    return slide


# prevent further operations on slide handle after it is closed
def _check_close(_result: Any, _func: Any, args: tuple[_OpenSlide]) -> None:
    args[0].invalidate()


# wrap the handle returned when creating a cache
def _check_cache_create(result: int, _func: Any, _args: Any) -> _OpenSlideCache:
    return _OpenSlideCache(c_void_p(result))


# Convert returned byte array, if present, into a string
def _check_string(result: Any, func: _CTypesFunc[..., Any], _args: Any) -> Any:
    if func.restype is c_char_p and result is not None:
        return result.decode('UTF-8', 'replace')
    else:
        return result


# check if the library got into an error state after each library call
def _check_error(result: Any, func: Any, args: tuple[Any, ...]) -> Any:
    assert isinstance(args[0], _OpenSlide)
    err = get_error(args[0])
    if err is not None:
        raise OpenSlideError(err)
    return _check_string(result, func, args)


# Convert returned NULL-terminated char** into a list of strings
def _check_name_list(result: _Pointer[c_char_p], func: Any, args: Any) -> list[str]:
    _check_error(result, func, args)
    names = []
    for i in count():
        name = result[i]
        if not name:
            break
        names.append(name.decode('UTF-8', 'replace'))
    return names


class _FunctionUnavailable:
    '''Standin for a missing optional function.  Fails when called.'''

    def __init__(self, minimum_version: str):
        self._minimum_version = minimum_version
        # allow checking for availability without calling the function
        self.available = False

    def __call__(self, *_args: Any) -> Any:
        raise OpenSlideVersionError(self._minimum_version)


# gate runtime code that requires ParamSpec, Python 3.10+
if TYPE_CHECKING:
    _P = ParamSpec('_P')
    _T = TypeVar('_T', covariant=True)

    class _Func(Protocol[_P, _T]):
        available: bool

        def __call__(self, *args: _P.args) -> _T: ...  # type: ignore[valid-type]

    class _CTypesFunc(_Func[_P, _T]):
        restype: type | None
        argtypes: list[type]
        errcheck: _ErrCheck

    _ErrCheck: TypeAlias = (
        Callable[[Any, _CTypesFunc[..., Any], tuple[Any, ...]], Any] | None
    )


# resolve and return an OpenSlide function with the specified properties
def _func(
    name: str,
    restype: type | None,
    argtypes: list[type],
    errcheck: _ErrCheck = _check_error,
    minimum_version: str | None = None,
) -> _Func[_P, _T]:
    try:
        func: _CTypesFunc[_P, _T] = getattr(_lib, name)
    except AttributeError:
        if minimum_version is None:
            raise
        else:
            return _FunctionUnavailable(minimum_version)
    func.argtypes = argtypes
    func.restype = restype
    if errcheck is not None:
        func.errcheck = errcheck
    func.available = True
    return func


def _wraps_funcs(
    wrapped: list[_Func[..., Any]],
) -> Callable[[Callable[_P, _T]], _Func[_P, _T]]:
    def decorator(fn: Callable[_P, _T]) -> _Func[_P, _T]:
        if TYPE_CHECKING:
            # requires ParamSpec, Python 3.10+
            f = cast(_Func[_P, _T], fn)
        else:
            f = fn
        f.available = True
        for w in wrapped:
            f.available = f.available and w.available
        return f

    return decorator


try:
    detect_vendor: _Func[[Filename], str] = _func(
        'openslide_detect_vendor', c_char_p, [_filename_p], _check_string
    )
except AttributeError:
    raise OpenSlideVersionError('3.4.0')

open: _Func[[Filename], _OpenSlide] = _func(
    'openslide_open', c_void_p, [_filename_p], _check_open
)

close: _Func[[_OpenSlide], None] = _func(
    'openslide_close', None, [_OpenSlide], _check_close
)

get_level_count: _Func[[_OpenSlide], int] = _func(
    'openslide_get_level_count', c_int32, [_OpenSlide]
)

_get_level_dimensions: _Func[
    [_OpenSlide, int, _Pointer[c_int64], _Pointer[c_int64]], None
] = _func(
    'openslide_get_level_dimensions',
    None,
    [_OpenSlide, c_int32, POINTER(c_int64), POINTER(c_int64)],
)


@_wraps_funcs([_get_level_dimensions])
def get_level_dimensions(slide: _OpenSlide, level: int) -> tuple[int, int]:
    w, h = c_int64(), c_int64()
    _get_level_dimensions(slide, level, byref(w), byref(h))
    return w.value, h.value


get_level_downsample: _Func[[_OpenSlide, int], float] = _func(
    'openslide_get_level_downsample', c_double, [_OpenSlide, c_int32]
)

get_best_level_for_downsample: _Func[[_OpenSlide, float], int] = _func(
    'openslide_get_best_level_for_downsample', c_int32, [_OpenSlide, c_double]
)

_read_region: _Func[[_OpenSlide, _Pointer[c_uint32], int, int, int, int, int], None] = (
    _func(
        'openslide_read_region',
        None,
        [_OpenSlide, POINTER(c_uint32), c_int64, c_int64, c_int32, c_int64, c_int64],
    )
)


@_wraps_funcs([_read_region])
def read_region(
    slide: _OpenSlide, x: int, y: int, level: int, w: int, h: int
) -> Image.Image:
    if w < 0 or h < 0:
        # OpenSlide would catch this, but not before we tried to allocate
        # a negative-size buffer
        raise OpenSlideError(
            "negative width (%d) or negative height (%d) not allowed" % (w, h)
        )
    if w == 0 or h == 0:
        # Image.frombuffer() would raise an exception
        return Image.new('RGBA', (w, h))
    buf = (w * h * c_uint32)()
    _read_region(slide, buf, x, y, level, w, h)
    return _load_image(buf, (w, h))


get_icc_profile_size: _Func[[_OpenSlide], int] = _func(
    'openslide_get_icc_profile_size',
    c_int64,
    [_OpenSlide],
    minimum_version='4.0.0',
)

_read_icc_profile: _Func[[_OpenSlide, _Pointer[c_char]], None] = _func(
    'openslide_read_icc_profile',
    None,
    [_OpenSlide, POINTER(c_char)],
    minimum_version='4.0.0',
)


@_wraps_funcs([get_icc_profile_size, _read_icc_profile])
def read_icc_profile(slide: _OpenSlide) -> bytes | None:
    size = get_icc_profile_size(slide)
    if size == 0:
        return None
    buf = (size * c_char)()
    _read_icc_profile(slide, buf)
    return buf.raw


get_error: _Func[[_OpenSlide], str] = _func(
    'openslide_get_error', c_char_p, [_OpenSlide], _check_string
)

get_property_names: _Func[[_OpenSlide], list[str]] = _func(
    'openslide_get_property_names', POINTER(c_char_p), [_OpenSlide], _check_name_list
)

get_property_value: _Func[[_OpenSlide, str | bytes], str] = _func(
    'openslide_get_property_value', c_char_p, [_OpenSlide, _utf8_p]
)

get_associated_image_names: _Func[[_OpenSlide], list[str]] = _func(
    'openslide_get_associated_image_names',
    POINTER(c_char_p),
    [_OpenSlide],
    _check_name_list,
)

_get_associated_image_dimensions: _Func[
    [_OpenSlide, str | bytes, _Pointer[c_int64], _Pointer[c_int64]], None
] = _func(
    'openslide_get_associated_image_dimensions',
    None,
    [_OpenSlide, _utf8_p, POINTER(c_int64), POINTER(c_int64)],
)


@_wraps_funcs([_get_associated_image_dimensions])
def get_associated_image_dimensions(
    slide: _OpenSlide, name: str | bytes
) -> tuple[int, int]:
    w, h = c_int64(), c_int64()
    _get_associated_image_dimensions(slide, name, byref(w), byref(h))
    return w.value, h.value


_read_associated_image: _Func[[_OpenSlide, str | bytes, _Pointer[c_uint32]], None] = (
    _func(
        'openslide_read_associated_image',
        None,
        [_OpenSlide, _utf8_p, POINTER(c_uint32)],
    )
)


@_wraps_funcs([get_associated_image_dimensions, _read_associated_image])
def read_associated_image(slide: _OpenSlide, name: str | bytes) -> Image.Image:
    w, h = get_associated_image_dimensions(slide, name)
    buf = (w * h * c_uint32)()
    _read_associated_image(slide, name, buf)
    return _load_image(buf, (w, h))


get_associated_image_icc_profile_size: _Func[[_OpenSlide, str | bytes], int] = _func(
    'openslide_get_associated_image_icc_profile_size',
    c_int64,
    [_OpenSlide, _utf8_p],
    minimum_version='4.0.0',
)

_read_associated_image_icc_profile: _Func[
    [_OpenSlide, str | bytes, _Pointer[c_char]], None
] = _func(
    'openslide_read_associated_image_icc_profile',
    None,
    [_OpenSlide, _utf8_p, POINTER(c_char)],
    minimum_version='4.0.0',
)


@_wraps_funcs(
    [get_associated_image_icc_profile_size, _read_associated_image_icc_profile]
)
def read_associated_image_icc_profile(
    slide: _OpenSlide, name: str | bytes
) -> bytes | None:
    size = get_associated_image_icc_profile_size(slide, name)
    if size == 0:
        return None
    buf = (size * c_char)()
    _read_associated_image_icc_profile(slide, name, buf)
    return buf.raw


get_version: _Func[[], str] = _func(
    'openslide_get_version', c_char_p, [], _check_string
)

cache_create: _Func[[int], _OpenSlideCache] = _func(
    'openslide_cache_create',
    c_void_p,
    [_size_t],
    _check_cache_create,
    minimum_version='4.0.0',
)

set_cache: _Func[[_OpenSlide, _OpenSlideCache], None] = _func(
    'openslide_set_cache',
    None,
    [_OpenSlide, _OpenSlideCache],
    minimum_version='4.0.0',
)

cache_release: _Func[[_OpenSlideCache], None] = _func(
    'openslide_cache_release', None, [_OpenSlideCache], None, minimum_version='4.0.0'
)



#!/usr/bin/env python
#
# deepzoom_multiserver - Example web application for viewing multiple slides
#
# Copyright (c) 2010-2015 Carnegie Mellon University
# Copyright (c) 2021-2024 Benjamin Gilbert
#
# This library is free software; you can redistribute it and/or modify it
# under the terms of version 2.1 of the GNU Lesser General Public License
# as published by the Free Software Foundation.
#
# This library is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public
# License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this library.  If not, see <https://www.gnu.org/licenses/>.
#

from __future__ import annotations

from argparse import ArgumentParser
import base64
from collections import OrderedDict
from collections.abc import Callable
from io import BytesIO
import os
from pathlib import Path, PurePath
from threading import Lock
from typing import TYPE_CHECKING, Any, Literal
import zlib

from PIL import Image, ImageCms
from flask import Flask, Response, abort, make_response, render_template, url_for

if TYPE_CHECKING:
    # Python 3.10+
    from typing import TypeAlias

if os.name == 'nt':
    _dll_path = os.getenv('OPENSLIDE_PATH')
    if _dll_path is not None:
        with os.add_dll_directory(_dll_path):  # type: ignore[attr-defined,unused-ignore]  # noqa: E501
            import openslide
    else:
        import openslide
else:
    import openslide

from openslide import OpenSlide, OpenSlideCache, OpenSlideError, OpenSlideVersionError
from openslide.deepzoom import DeepZoomGenerator

# Optimized sRGB v2 profile, CC0-1.0 license
# https://github.com/saucecontrol/Compact-ICC-Profiles/blob/bdd84663/profiles/sRGB-v2-micro.icc
# ImageCms.createProfile() generates a v4 profile and Firefox has problems
# with those: https://littlecms.com/blog/2020/09/09/browser-check/
SRGB_PROFILE_BYTES = zlib.decompress(
    base64.b64decode(
        'eNpjYGA8kZOcW8wkwMCQm1dSFOTupBARGaXA/oiBmUGEgZOBj0E2Mbm4wDfYLYQBCIoT'
        'y4uTS4pyGFDAt2sMjCD6sm5GYl7K3IkMtg4NG2wdSnQa5y1V6mPADzhTUouTgfQHII5P'
        'LigqYWBg5AGyecpLCkBsCSBbpAjoKCBbB8ROh7AdQOwkCDsErCYkyBnIzgCyE9KR2ElI'
        'bKhdIMBaCvQsskNKUitKQLSzswEDKAwgop9DwH5jFDuJEMtfwMBg8YmBgbkfIZY0jYFh'
        'eycDg8QthJgKUB1/KwPDtiPJpUVlUGu0gLiG4QfjHKZS5maWk2x+HEJcEjxJfF8Ez4t8'
        'k8iS0VNwVlmjmaVXZ/zacrP9NbdwX7OQshjxFNmcttKwut4OnUlmc1Yv79l0e9/MU8ev'
        'pz4p//jz/38AR4Nk5Q=='
    )
)
SRGB_PROFILE = ImageCms.getOpenProfile(BytesIO(SRGB_PROFILE_BYTES))

if TYPE_CHECKING:
    ColorMode: TypeAlias = Literal[
        'default',
        'absolute-colorimetric',
        'perceptual',
        'relative-colorimetric',
        'saturation',
        'embed',
        'ignore',
    ]
    Transform: TypeAlias = Callable[[Image.Image], None]


class DeepZoomMultiServer(Flask):
    basedir: Path
    cache: _SlideCache


class AnnotatedDeepZoomGenerator(DeepZoomGenerator):
    filename: str
    mpp: float
    transform: Transform


def create_app(
    config: dict[str, Any] | None = None,
    config_file: Path | None = None,
) -> Flask:
    # Create and configure app
    app = DeepZoomMultiServer(__name__)
    app.config.from_mapping(
        SLIDE_DIR='.',
        SLIDE_CACHE_SIZE=10,
        SLIDE_TILE_CACHE_MB=128,
        DEEPZOOM_FORMAT='jpeg',
        DEEPZOOM_TILE_SIZE=254,
        DEEPZOOM_OVERLAP=1,
        DEEPZOOM_LIMIT_BOUNDS=True,
        DEEPZOOM_TILE_QUALITY=75,
        DEEPZOOM_COLOR_MODE='default',
    )
    app.config.from_envvar('DEEPZOOM_MULTISERVER_SETTINGS', silent=True)
    if config_file is not None:
        app.config.from_pyfile(config_file)
    if config is not None:
        app.config.from_mapping(config)

    # Set up cache
    app.basedir = Path(app.config['SLIDE_DIR']).resolve(strict=True)
    config_map = {
        'DEEPZOOM_TILE_SIZE': 'tile_size',
        'DEEPZOOM_OVERLAP': 'overlap',
        'DEEPZOOM_LIMIT_BOUNDS': 'limit_bounds',
    }
    opts = {v: app.config[k] for k, v in config_map.items()}
    app.cache = _SlideCache(
        app.config['SLIDE_CACHE_SIZE'],
        app.config['SLIDE_TILE_CACHE_MB'],
        opts,
        app.config['DEEPZOOM_COLOR_MODE'],
    )

    # Helper functions
    def get_slide(user_path: PurePath) -> AnnotatedDeepZoomGenerator:
        try:
            path = (app.basedir / user_path).resolve(strict=True)
        except OSError:
            # Does not exist
            abort(404)
        if path.parts[: len(app.basedir.parts)] != app.basedir.parts:
            # Directory traversal
            abort(404)
        try:
            slide = app.cache.get(path)
            slide.filename = path.name
            return slide
        except OpenSlideError:
            abort(404)

    # Set up routes
    @app.route('/')
    def index() -> str:
        return render_template('files.html', root_dir=_Directory(app.basedir))

    @app.route('/<path:path>')
    def slide(path: str) -> str:
        slide = get_slide(PurePath(path))
        slide_url = url_for('dzi', path=path)
        return render_template(
            'slide-fullpage.html',
            slide_url=slide_url,
            slide_filename=slide.filename,
            slide_mpp=slide.mpp,
        )

    @app.route('/<path:path>.dzi')
    def dzi(path: str) -> Response:
        slide = get_slide(PurePath(path))
        format = app.config['DEEPZOOM_FORMAT']
        resp = make_response(slide.get_dzi(format))
        resp.mimetype = 'application/xml'
        return resp

    @app.route('/<path:path>_files/<int:level>/<int:col>_<int:row>.<format>')
    def tile(path: str, level: int, col: int, row: int, format: str) -> Response:
        slide = get_slide(PurePath(path))
        format = format.lower()
        if format != 'jpeg' and format != 'png':
            # Not supported by Deep Zoom
            abort(404)
        try:
            tile = slide.get_tile(level, (col, row))
        except ValueError:
            # Invalid level or coordinates
            abort(404)
        slide.transform(tile)
        buf = BytesIO()
        tile.save(
            buf,
            format,
            quality=app.config['DEEPZOOM_TILE_QUALITY'],
            icc_profile=tile.info.get('icc_profile'),
        )
        resp = make_response(buf.getvalue())
        resp.mimetype = 'image/%s' % format
        return resp

    return app


class _SlideCache:
    def __init__(
        self,
        cache_size: int,
        tile_cache_mb: int,
        dz_opts: dict[str, Any],
        color_mode: ColorMode,
    ):
        self.cache_size = cache_size
        self.dz_opts = dz_opts
        self.color_mode = color_mode
        self._lock = Lock()
        self._cache: OrderedDict[Path, AnnotatedDeepZoomGenerator] = OrderedDict()
        # Share a single tile cache among all slide handles, if supported
        try:
            self._tile_cache: OpenSlideCache | None = OpenSlideCache(
                tile_cache_mb * 1024 * 1024
            )
        except OpenSlideVersionError:
            self._tile_cache = None

    def get(self, path: Path) -> AnnotatedDeepZoomGenerator:
        with self._lock:
            if path in self._cache:
                # Move to end of LRU
                slide = self._cache.pop(path)
                self._cache[path] = slide
                return slide

        osr = OpenSlide(path)
        if self._tile_cache is not None:
            osr.set_cache(self._tile_cache)
        slide = AnnotatedDeepZoomGenerator(osr, **self.dz_opts)
        try:
            mpp_x = osr.properties[openslide.PROPERTY_NAME_MPP_X]
            mpp_y = osr.properties[openslide.PROPERTY_NAME_MPP_Y]
            slide.mpp = (float(mpp_x) + float(mpp_y)) / 2
        except (KeyError, ValueError):
            slide.mpp = 0
        slide.transform = self._get_transform(osr)

        with self._lock:
            if path not in self._cache:
                if len(self._cache) == self.cache_size:
                    self._cache.popitem(last=False)
                self._cache[path] = slide
        return slide

    def _get_transform(self, image: OpenSlide) -> Transform:
        if image.color_profile is None:
            return lambda img: None
        mode = self.color_mode
        if mode == 'ignore':
            # drop ICC profile from tiles
            return lambda img: img.info.pop('icc_profile')
        elif mode == 'embed':
            # embed ICC profile in tiles
            return lambda img: None
        elif mode == 'default':
            intent = ImageCms.Intent(ImageCms.getDefaultIntent(image.color_profile))
        elif mode == 'absolute-colorimetric':
            intent = ImageCms.Intent.ABSOLUTE_COLORIMETRIC
        elif mode == 'relative-colorimetric':
            intent = ImageCms.Intent.RELATIVE_COLORIMETRIC
        elif mode == 'perceptual':
            intent = ImageCms.Intent.PERCEPTUAL
        elif mode == 'saturation':
            intent = ImageCms.Intent.SATURATION
        else:
            raise ValueError(f'Unknown color mode {mode}')
        transform = ImageCms.buildTransform(
            image.color_profile,
            SRGB_PROFILE,
            'RGB',
            'RGB',
            intent,
            ImageCms.Flags(0),
        )

        def xfrm(img: Image.Image) -> None:
            ImageCms.applyTransform(img, transform, True)
            # Some browsers assume we intend the display's color space if we
            # don't embed the profile.  Pillow's serialization is larger, so
            # use ours.
            img.info['icc_profile'] = SRGB_PROFILE_BYTES

        return xfrm


class _Directory:
    _DEFAULT_RELPATH = PurePath('.')

    def __init__(self, basedir: Path, relpath: PurePath = _DEFAULT_RELPATH):
        self.name = relpath.name
        self.children: list[_Directory | _SlideFile] = []
        for cur_path in sorted((basedir / relpath).iterdir()):
            cur_relpath = relpath / cur_path.name
            if cur_path.is_dir():
                cur_dir = _Directory(basedir, cur_relpath)
                if cur_dir.children:
                    self.children.append(cur_dir)
            elif OpenSlide.detect_format(cur_path):
                self.children.append(_SlideFile(cur_relpath))


class _SlideFile:
    def __init__(self, relpath: PurePath):
        self.name = relpath.name
        self.url_path = relpath.as_posix()


if __name__ == '__main__':
    parser = ArgumentParser(usage='%(prog)s [options] [SLIDE-DIRECTORY]')
    parser.add_argument(
        '-B',
        '--ignore-bounds',
        dest='DEEPZOOM_LIMIT_BOUNDS',
        default=True,
        action='store_false',
        help='display entire scan area',
    )
    parser.add_argument(
        '--color-mode',
        dest='DEEPZOOM_COLOR_MODE',
        choices=[
            'default',
            'absolute-colorimetric',
            'perceptual',
            'relative-colorimetric',
            'saturation',
            'embed',
            'ignore',
        ],
        default='default',
        help=(
            'convert tiles to sRGB using default rendering intent of ICC '
            'profile, or specified rendering intent; or embed original '
            'ICC profile; or ignore ICC profile (compat) [default]'
        ),
    )
    parser.add_argument(
        '-c', '--config', metavar='FILE', type=Path, dest='config', help='config file'
    )
    parser.add_argument(
        '-d',
        '--debug',
        dest='DEBUG',
        action='store_true',
        help='run in debugging mode (insecure)',
    )
    parser.add_argument(
        '-e',
        '--overlap',
        metavar='PIXELS',
        dest='DEEPZOOM_OVERLAP',
        type=int,
        help='overlap of adjacent tiles [1]',
    )
    parser.add_argument(
        '-f',
        '--format',
        dest='DEEPZOOM_FORMAT',
        choices=['jpeg', 'png'],
        help='image format for tiles [jpeg]',
    )
    parser.add_argument(
        '-l',
        '--listen',
        metavar='ADDRESS',
        dest='host',
        default='127.0.0.1',
        help='address to listen on [127.0.0.1]',
    )
    parser.add_argument(
        '-p',
        '--port',
        metavar='PORT',
        dest='port',
        type=int,
        default=5000,
        help='port to listen on [5000]',
    )
    parser.add_argument(
        '-Q',
        '--quality',
        metavar='QUALITY',
        dest='DEEPZOOM_TILE_QUALITY',
        type=int,
        help='JPEG compression quality [75]',
    )
    parser.add_argument(
        '-s',
        '--size',
        metavar='PIXELS',
        dest='DEEPZOOM_TILE_SIZE',
        type=int,
        help='tile size [254]',
    )
    parser.add_argument(
        'SLIDE_DIR',
        metavar='SLIDE-DIRECTORY',
        type=Path,
        nargs='?',
        help='slide directory',
    )

    args = parser.parse_args()
    config = {}
    config_file = args.config
    # Set only those settings specified on the command line
    for k in dir(args):
        v = getattr(args, k)
        if not k.startswith('_') and v is not None:
            config[k] = v
    app = create_app(config, config_file)

    app.run(host=args.host, port=args.port, threaded=True)



#
# openslide-python - Python bindings for the OpenSlide library
#
# Copyright (c) 2016-2023 Benjamin Gilbert
#
# This library is free software; you can redistribute it and/or modify it
# under the terms of version 2.1 of the GNU Lesser General Public License
# as published by the Free Software Foundation.
#
# This library is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public
# License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this library.  If not, see <https://www.gnu.org/licenses/>.
#

from __future__ import annotations

import ctypes
import unittest

from common import file_path

from openslide import ImageSlide, OpenSlide, lowlevel, open_slide


class TestLibrary(unittest.TestCase):
    def test_open_slide(self) -> None:
        with open_slide(file_path('boxes.tiff')) as osr:
            self.assertTrue(isinstance(osr, OpenSlide))
        with open_slide(file_path('boxes.png')) as osr:
            self.assertTrue(isinstance(osr, ImageSlide))

    def test_lowlevel_available(self) -> None:
        '''Ensure all exported functions have an 'available' attribute.'''
        for name in dir(lowlevel):
            attr = getattr(lowlevel, name)
            # ignore classes and unexported functions
            if name.startswith('_') or name[0].isupper():
                continue
            # ignore __future__ imports
            if getattr(attr, '__module__', None) == '__future__':
                continue
            # ignore random imports
            if hasattr(ctypes, name) or name in ('count', 'os', 'platform'):
                continue
            self.assertTrue(
                hasattr(attr, 'available'),
                f'"{name}" missing "available" attribute',
            )



#!/usr/bin/env python
#
# deepzoom_tile - Convert whole-slide images to Deep Zoom format
#
# Copyright (c) 2010-2015 Carnegie Mellon University
# Copyright (c) 2022-2024 Benjamin Gilbert
#
# This library is free software; you can redistribute it and/or modify it
# under the terms of version 2.1 of the GNU Lesser General Public License
# as published by the Free Software Foundation.
#
# This library is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public
# License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this library.  If not, see <https://www.gnu.org/licenses/>.
#

"""An example program to generate a Deep Zoom directory tree from a slide."""

from __future__ import annotations

from argparse import ArgumentParser
import base64
from collections.abc import Callable
from io import BytesIO
import json
from multiprocessing import JoinableQueue, Process
import multiprocessing.queues
import os
from pathlib import Path
import re
import shutil
import sys
from typing import TYPE_CHECKING, Literal
from unicodedata import normalize
import zlib

from PIL import Image, ImageCms

if TYPE_CHECKING:
    # Python 3.10+
    from typing import TypeAlias

if os.name == 'nt':
    _dll_path = os.getenv('OPENSLIDE_PATH')
    if _dll_path is not None:
        with os.add_dll_directory(_dll_path):  # type: ignore[attr-defined,unused-ignore]  # noqa: E501
            import openslide
    else:
        import openslide
else:
    import openslide

from openslide import AbstractSlide, ImageSlide, open_slide
from openslide.deepzoom import DeepZoomGenerator

VIEWER_SLIDE_NAME = 'slide'

# Optimized sRGB v2 profile, CC0-1.0 license
# https://github.com/saucecontrol/Compact-ICC-Profiles/blob/bdd84663/profiles/sRGB-v2-micro.icc
# ImageCms.createProfile() generates a v4 profile and Firefox has problems
# with those: https://littlecms.com/blog/2020/09/09/browser-check/
SRGB_PROFILE_BYTES = zlib.decompress(
    base64.b64decode(
        'eNpjYGA8kZOcW8wkwMCQm1dSFOTupBARGaXA/oiBmUGEgZOBj0E2Mbm4wDfYLYQBCIoT'
        'y4uTS4pyGFDAt2sMjCD6sm5GYl7K3IkMtg4NG2wdSnQa5y1V6mPADzhTUouTgfQHII5P'
        'LigqYWBg5AGyecpLCkBsCSBbpAjoKCBbB8ROh7AdQOwkCDsErCYkyBnIzgCyE9KR2ElI'
        'bKhdIMBaCvQsskNKUitKQLSzswEDKAwgop9DwH5jFDuJEMtfwMBg8YmBgbkfIZY0jYFh'
        'eycDg8QthJgKUB1/KwPDtiPJpUVlUGu0gLiG4QfjHKZS5maWk2x+HEJcEjxJfF8Ez4t8'
        'k8iS0VNwVlmjmaVXZ/zacrP9NbdwX7OQshjxFNmcttKwut4OnUlmc1Yv79l0e9/MU8ev'
        'pz4p//jz/38AR4Nk5Q=='
    )
)
SRGB_PROFILE = ImageCms.getOpenProfile(BytesIO(SRGB_PROFILE_BYTES))

if TYPE_CHECKING:
    ColorMode: TypeAlias = Literal[
        'default',
        'absolute-colorimetric',
        'perceptual',
        'relative-colorimetric',
        'saturation',
        'embed',
        'ignore',
    ]
    TileQueue: TypeAlias = multiprocessing.queues.JoinableQueue[
        tuple[str | None, int, tuple[int, int], Path] | None
    ]
    Transform: TypeAlias = Callable[[Image.Image], None]


class TileWorker(Process):
    """A child process that generates and writes tiles."""

    def __init__(
        self,
        queue: TileQueue,
        slidepath: Path,
        tile_size: int,
        overlap: int,
        limit_bounds: bool,
        quality: int,
        color_mode: ColorMode,
    ):
        Process.__init__(self, name='TileWorker')
        self.daemon = True
        self._queue = queue
        self._slidepath = slidepath
        self._tile_size = tile_size
        self._overlap = overlap
        self._limit_bounds = limit_bounds
        self._quality = quality
        self._color_mode = color_mode
        self._slide: AbstractSlide | None = None

    def run(self) -> None:
        self._slide = open_slide(self._slidepath)
        last_associated = None
        dz, transform = self._get_dz_and_transform()
        while True:
            data = self._queue.get()
            if data is None:
                self._queue.task_done()
                break
            associated, level, address, outfile = data
            if last_associated != associated:
                dz, transform = self._get_dz_and_transform(associated)
                last_associated = associated
            tile = dz.get_tile(level, address)
            transform(tile)
            tile.save(
                outfile, quality=self._quality, icc_profile=tile.info.get('icc_profile')
            )
            self._queue.task_done()

    def _get_dz_and_transform(
        self, associated: str | None = None
    ) -> tuple[DeepZoomGenerator, Transform]:
        assert self._slide is not None
        if associated is not None:
            image: AbstractSlide = ImageSlide(self._slide.associated_images[associated])
        else:
            image = self._slide
        dz = DeepZoomGenerator(
            image, self._tile_size, self._overlap, limit_bounds=self._limit_bounds
        )
        return dz, self._get_transform(image)

    def _get_transform(self, image: AbstractSlide) -> Transform:
        if image.color_profile is None:
            return lambda img: None
        mode = self._color_mode
        if mode == 'ignore':
            # drop ICC profile from tiles
            return lambda img: img.info.pop('icc_profile')
        elif mode == 'embed':
            # embed ICC profile in tiles
            return lambda img: None
        elif mode == 'default':
            intent = ImageCms.Intent(ImageCms.getDefaultIntent(image.color_profile))
        elif mode == 'absolute-colorimetric':
            intent = ImageCms.Intent.ABSOLUTE_COLORIMETRIC
        elif mode == 'relative-colorimetric':
            intent = ImageCms.Intent.RELATIVE_COLORIMETRIC
        elif mode == 'perceptual':
            intent = ImageCms.Intent.PERCEPTUAL
        elif mode == 'saturation':
            intent = ImageCms.Intent.SATURATION
        else:
            raise ValueError(f'Unknown color mode {mode}')
        transform = ImageCms.buildTransform(
            image.color_profile,
            SRGB_PROFILE,
            'RGB',
            'RGB',
            intent,
            ImageCms.Flags(0),
        )

        def xfrm(img: Image.Image) -> None:
            ImageCms.applyTransform(img, transform, True)
            # Some browsers assume we intend the display's color space if we
            # don't embed the profile.  Pillow's serialization is larger, so
            # use ours.
            img.info['icc_profile'] = SRGB_PROFILE_BYTES

        return xfrm


class DeepZoomImageTiler:
    """Handles generation of tiles and metadata for a single image."""

    def __init__(
        self,
        dz: DeepZoomGenerator,
        basename: Path,
        format: str,
        associated: str | None,
        queue: TileQueue,
    ):
        self._dz = dz
        self._basename = basename
        self._format = format
        self._associated = associated
        self._queue = queue
        self._processed = 0

    def run(self) -> None:
        self._write_tiles()
        self._write_dzi()

    def _write_tiles(self) -> None:
        for level in range(self._dz.level_count):
            tiledir = self._basename.with_name(self._basename.name + '_files') / str(
                level
            )
            tiledir.mkdir(parents=True, exist_ok=True)
            cols, rows = self._dz.level_tiles[level]
            for row in range(rows):
                for col in range(cols):
                    tilename = tiledir / f'{col}_{row}.{self._format}'
                    if not tilename.exists():
                        self._queue.put((self._associated, level, (col, row), tilename))
                    self._tile_done()

    def _tile_done(self) -> None:
        self._processed += 1
        count, total = self._processed, self._dz.tile_count
        if count % 100 == 0 or count == total:
            print(
                "Tiling %s: wrote %d/%d tiles"
                % (self._associated or 'slide', count, total),
                end='\r',
                file=sys.stderr,
            )
            if count == total:
                print(file=sys.stderr)

    def _write_dzi(self) -> None:
        with self._basename.with_name(self._basename.name + '.dzi').open('w') as fh:
            fh.write(self.get_dzi())

    def get_dzi(self) -> str:
        return self._dz.get_dzi(self._format)


class DeepZoomStaticTiler:
    """Handles generation of tiles and metadata for all images in a slide."""

    def __init__(
        self,
        slidepath: Path,
        basename: Path,
        format: str,
        tile_size: int,
        overlap: int,
        limit_bounds: bool,
        quality: int,
        color_mode: ColorMode,
        workers: int,
        with_viewer: bool,
    ):
        if with_viewer:
            # Check extra dependency before doing a bunch of work
            import jinja2  # noqa: F401  module-imported-but-unused
        self._slide = open_slide(slidepath)
        self._basename = basename
        self._format = format
        self._tile_size = tile_size
        self._overlap = overlap
        self._limit_bounds = limit_bounds
        self._queue: TileQueue = JoinableQueue(2 * workers)
        self._workers = workers
        self._color_mode = color_mode
        self._with_viewer = with_viewer
        self._dzi_data: dict[str, str] = {}
        for _i in range(workers):
            TileWorker(
                self._queue,
                slidepath,
                tile_size,
                overlap,
                limit_bounds,
                quality,
                color_mode,
            ).start()

    def run(self) -> None:
        self._run_image()
        if self._with_viewer:
            for name in self._slide.associated_images:
                self._run_image(name)
            self._write_html()
            self._write_static()
        self._shutdown()

    def _run_image(self, associated: str | None = None) -> None:
        """Run a single image from self._slide."""
        if associated is None:
            image = self._slide
            if self._with_viewer:
                basename = self._basename / VIEWER_SLIDE_NAME
            else:
                basename = self._basename
        else:
            image = ImageSlide(self._slide.associated_images[associated])
            basename = self._basename / self._slugify(associated)
        dz = DeepZoomGenerator(
            image, self._tile_size, self._overlap, limit_bounds=self._limit_bounds
        )
        tiler = DeepZoomImageTiler(dz, basename, self._format, associated, self._queue)
        tiler.run()
        self._dzi_data[self._url_for(associated)] = tiler.get_dzi()

    def _url_for(self, associated: str | None) -> str:
        if associated is None:
            base = VIEWER_SLIDE_NAME
        else:
            base = self._slugify(associated)
        return '%s.dzi' % base

    def _write_html(self) -> None:
        import jinja2

        # https://docs.python.org/3/reference/import.html#main-spec
        if __spec__ is not None:
            # We're running from a module (e.g. "python -m deepzoom_tile")
            # so load templates from the containing package.
            loader = jinja2.PackageLoader('__main__')
        else:
            # We're not running from a module (e.g. "python deepzoom_tile.py")
            # so PackageLoader('__main__') doesn't work in jinja2 3.x.
            # Load templates directly from the filesystem.
            loader = jinja2.FileSystemLoader([Path(__file__).parent / 'templates'])
        env = jinja2.Environment(loader=loader, autoescape=True)
        template = env.get_template('slide-multipane.html')
        associated_urls = {n: self._url_for(n) for n in self._slide.associated_images}
        try:
            mpp_x = self._slide.properties[openslide.PROPERTY_NAME_MPP_X]
            mpp_y = self._slide.properties[openslide.PROPERTY_NAME_MPP_Y]
            mpp = (float(mpp_x) + float(mpp_y)) / 2
        except (KeyError, ValueError):
            mpp = 0
        # Embed the dzi metadata in the HTML to work around Chrome's
        # refusal to allow XmlHttpRequest from file:///, even when
        # the originating page is also a file:///
        data = template.render(
            slide_url=self._url_for(None),
            slide_mpp=mpp,
            associated=associated_urls,
            properties=self._slide.properties,
            dzi_data=json.dumps(self._dzi_data),
        )
        with open(self._basename / 'index.html', 'w') as fh:
            fh.write(data)

    def _write_static(self) -> None:
        basesrc = Path(__file__).absolute().parent / 'static'
        basedst = self._basename / 'static'
        self._copydir(basesrc, basedst)
        self._copydir(basesrc / 'images', basedst / 'images')

    def _copydir(self, src: Path, dest: Path) -> None:
        dest.mkdir(parents=True, exist_ok=True)
        for srcpath in src.iterdir():
            if srcpath.is_file():
                shutil.copy(srcpath, dest / srcpath.name)

    @classmethod
    def _slugify(cls, text: str) -> str:
        text = normalize('NFKD', text.lower()).encode('ascii', 'ignore').decode()
        return re.sub('[^a-z0-9]+', '_', text)

    def _shutdown(self) -> None:
        for _i in range(self._workers):
            self._queue.put(None)
        self._queue.join()


if __name__ == '__main__':
    try:
        # Python 3.13+
        available_cpus = os.process_cpu_count()  # type: ignore[attr-defined]
    except AttributeError:
        try:
            # Linux
            available_cpus = len(
                os.sched_getaffinity(0)  # type: ignore[attr-defined,unused-ignore]
            )
        except AttributeError:
            # default
            available_cpus = 4

    parser = ArgumentParser(usage='%(prog)s [options] <SLIDE>')
    parser.add_argument(
        '-B',
        '--ignore-bounds',
        dest='limit_bounds',
        default=True,
        action='store_false',
        help='display entire scan area',
    )
    parser.add_argument(
        '--color-mode',
        dest='color_mode',
        choices=[
            'default',
            'absolute-colorimetric',
            'perceptual',
            'relative-colorimetric',
            'saturation',
            'embed',
            'ignore',
        ],
        default='default',
        help=(
            'convert tiles to sRGB using default rendering intent of ICC '
            'profile, or specified rendering intent; or embed original '
            'ICC profile; or ignore ICC profile (compat) [default]'
        ),
    )
    parser.add_argument(
        '-e',
        '--overlap',
        metavar='PIXELS',
        dest='overlap',
        type=int,
        default=1,
        help='overlap of adjacent tiles [1]',
    )
    parser.add_argument(
        '-f',
        '--format',
        dest='format',
        default='jpeg',
        choices=['jpeg', 'png'],
        help='image format for tiles [jpeg]',
    )
    parser.add_argument(
        '-j',
        '--jobs',
        metavar='COUNT',
        dest='workers',
        type=int,
        default=available_cpus,
        help=f'number of worker processes to start [{available_cpus}]',
    )
    parser.add_argument(
        '-o',
        '--output',
        metavar='NAME',
        type=Path,
        dest='basename',
        help='base name of output file',
    )
    parser.add_argument(
        '-Q',
        '--quality',
        metavar='QUALITY',
        dest='quality',
        type=int,
        default=90,
        help='JPEG compression quality [90]',
    )
    parser.add_argument(
        '-r',
        '--viewer',
        dest='with_viewer',
        action='store_true',
        help='generate directory tree with HTML viewer',
    )
    parser.add_argument(
        '-s',
        '--size',
        metavar='PIXELS',
        dest='tile_size',
        type=int,
        default=254,
        help='tile size [254]',
    )
    parser.add_argument(
        'slidepath',
        metavar='SLIDE',
        type=Path,
        help='slide file',
    )

    args = parser.parse_args()
    if args.basename is None:
        args.basename = Path(args.slidepath.stem)

    DeepZoomStaticTiler(
        args.slidepath,
        args.basename,
        args.format,
        args.tile_size,
        args.overlap,
        args.limit_bounds,
        args.quality,
        args.color_mode,
        args.workers,
        args.with_viewer,
    ).run()



#
# openslide-python - Python bindings for the OpenSlide library
#
# Copyright (c) 2016-2024 Benjamin Gilbert
#
# This library is free software; you can redistribute it and/or modify it
# under the terms of version 2.1 of the GNU Lesser General Public License
# as published by the Free Software Foundation.
#
# This library is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public
# License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this library.  If not, see <https://www.gnu.org/licenses/>.
#

from __future__ import annotations

from ctypes import ArgumentError
import re
import sys
from typing import Any
import unittest

from common import file_path

from openslide import (
    OpenSlide,
    OpenSlideCache,
    OpenSlideError,
    OpenSlideUnsupportedFormatError,
    lowlevel,
)


class TestCache(unittest.TestCase):
    @unittest.skipUnless(lowlevel.cache_create.available, "requires OpenSlide 4.0.0")
    def test_create_cache(self) -> None:
        OpenSlideCache(0)
        OpenSlideCache(1)
        OpenSlideCache(4 << 20)
        self.assertRaises(ArgumentError, lambda: OpenSlideCache(-1))
        self.assertRaises(
            ArgumentError, lambda: OpenSlideCache(1.3)  # type: ignore[arg-type]
        )


class TestSlideWithoutOpening(unittest.TestCase):
    def test_detect_format(self) -> None:
        self.assertTrue(OpenSlide.detect_format(file_path('__missing_file')) is None)
        self.assertTrue(OpenSlide.detect_format(file_path('../setup.py')) is None)
        self.assertEqual(
            OpenSlide.detect_format(file_path('boxes.tiff')), 'generic-tiff'
        )

    def test_open(self) -> None:
        self.assertRaises(
            OpenSlideUnsupportedFormatError, lambda: OpenSlide('__does_not_exist')
        )
        self.assertRaises(
            OpenSlideUnsupportedFormatError, lambda: OpenSlide('setup.py')
        )
        self.assertRaises(
            ArgumentError,
            lambda: OpenSlide(None),  # type: ignore[arg-type]
        )
        self.assertRaises(
            ArgumentError,
            lambda: OpenSlide(3),  # type: ignore[arg-type]
        )
        self.assertRaises(
            OpenSlideUnsupportedFormatError, lambda: OpenSlide('unopenable.tiff')
        )

    @unittest.skipUnless(
        sys.getfilesystemencoding() == 'utf-8',
        'Python filesystem encoding is not UTF-8',
    )
    def test_unicode_path(self) -> None:
        path = file_path('😐.svs')
        for arg in path, str(path):
            self.assertEqual(OpenSlide.detect_format(arg), 'aperio')
            self.assertEqual(OpenSlide(arg).dimensions, (16, 16))

    def test_unicode_path_bytes(self) -> None:
        arg = str(file_path('😐.svs')).encode('UTF-8')
        self.assertEqual(OpenSlide.detect_format(arg), 'aperio')
        self.assertEqual(OpenSlide(arg).dimensions, (16, 16))

    def test_operations_on_closed_handle(self) -> None:
        osr = OpenSlide(file_path('boxes.tiff'))
        props = osr.properties
        associated = osr.associated_images
        osr.close()
        self.assertRaises(ArgumentError, lambda: osr.read_region((0, 0), 0, (100, 100)))
        self.assertRaises(ArgumentError, lambda: osr.close())
        self.assertRaises(ArgumentError, lambda: props['openslide.vendor'])
        self.assertRaises(ArgumentError, lambda: associated['label'])

    def test_context_manager(self) -> None:
        osr = OpenSlide(file_path('boxes.tiff'))
        with osr:
            self.assertEqual(osr.level_count, 4)
        self.assertRaises(ArgumentError, lambda: osr.level_count)


class _Abstract:
    # nested class to prevent the test runner from finding it
    class SlideTest(unittest.TestCase):
        FILENAME: str | None = None

        def setUp(self) -> None:
            assert self.FILENAME is not None
            self.osr = OpenSlide(file_path(self.FILENAME))

        def tearDown(self) -> None:
            self.osr.close()


class TestSlide(_Abstract.SlideTest):
    FILENAME = 'boxes.tiff'

    def test_repr(self) -> None:
        self.assertEqual(repr(self.osr), 'OpenSlide(%r)' % file_path('boxes.tiff'))

    def test_basic_metadata(self) -> None:
        self.assertEqual(self.osr.level_count, 4)
        self.assertEqual(
            self.osr.level_dimensions, ((300, 250), (150, 125), (75, 62), (37, 31))
        )
        self.assertEqual(self.osr.dimensions, (300, 250))

        self.assertEqual(len(self.osr.level_downsamples), self.osr.level_count)
        self.assertEqual(self.osr.level_downsamples[0:2], (1, 2))
        self.assertAlmostEqual(self.osr.level_downsamples[2], 4, places=0)
        self.assertAlmostEqual(self.osr.level_downsamples[3], 8, places=0)

        self.assertEqual(self.osr.get_best_level_for_downsample(0.5), 0)
        self.assertEqual(self.osr.get_best_level_for_downsample(3), 1)
        self.assertEqual(self.osr.get_best_level_for_downsample(37), 3)

    def test_properties(self) -> None:
        self.assertEqual(self.osr.properties['openslide.vendor'], 'generic-tiff')
        self.assertRaises(KeyError, lambda: self.osr.properties['__does_not_exist'])
        # test __len__ and __iter__
        self.assertEqual(
            len([v for v in self.osr.properties]), len(self.osr.properties)
        )
        self.assertEqual(
            repr(self.osr.properties), '<_PropertyMap %r>' % dict(self.osr.properties)
        )

    @unittest.skipUnless(
        lowlevel.read_icc_profile.available, "requires OpenSlide 4.0.0"
    )
    def test_color_profile(self) -> None:
        assert self.osr.color_profile is not None  # for type inference
        self.assertEqual(self.osr.color_profile.profile.device_class, 'mntr')
        self.assertEqual(
            len(self.osr.read_region((0, 0), 0, (100, 100)).info['icc_profile']), 588
        )
        self.assertEqual(
            len(self.osr.get_thumbnail((100, 100)).info['icc_profile']), 588
        )

    def test_read_region(self) -> None:
        self.assertEqual(
            self.osr.read_region((-10, -10), 1, (400, 400)).size, (400, 400)
        )

    def test_read_region_size_dimension_zero(self) -> None:
        self.assertEqual(self.osr.read_region((0, 0), 1, (400, 0)).size, (400, 0))

    def test_read_region_bad_level(self) -> None:
        self.assertEqual(self.osr.read_region((0, 0), 4, (100, 100)).size, (100, 100))

    def test_read_region_bad_size(self) -> None:
        self.assertRaises(
            OpenSlideError, lambda: self.osr.read_region((0, 0), 1, (400, -5))
        )

    @unittest.skipIf(sys.maxsize < 1 << 32, '32-bit Python')
    # Disabled to avoid OOM killer on small systems, since the stdlib
    # doesn't provide a way to find out how much RAM we have
    def _test_read_region_2GB(self) -> None:
        self.assertEqual(
            self.osr.read_region((1000, 1000), 0, (32768, 16384)).size, (32768, 16384)
        )

    def test_thumbnail(self) -> None:
        self.assertEqual(self.osr.get_thumbnail((100, 100)).size, (100, 83))

    @unittest.skipUnless(lowlevel.cache_create.available, "requires OpenSlide 4.0.0")
    def test_set_cache(self) -> None:
        self.osr.set_cache(OpenSlideCache(64 << 10))
        self.assertEqual(self.osr.read_region((0, 0), 0, (400, 400)).size, (400, 400))
        self.assertRaises(
            TypeError, lambda: self.osr.set_cache(None)  # type: ignore[arg-type]
        )
        self.assertRaises(
            TypeError, lambda: self.osr.set_cache(3)  # type: ignore[arg-type]
        )


class TestAperioSlide(_Abstract.SlideTest):
    FILENAME = 'small.svs'

    def test_associated_images(self) -> None:
        self.assertEqual(self.osr.associated_images['thumbnail'].size, (16, 16))
        self.assertRaises(KeyError, lambda: self.osr.associated_images['__missing'])
        # test __len__ and __iter__
        self.assertEqual(
            len([v for v in self.osr.associated_images]),
            len(self.osr.associated_images),
        )

        def mangle_repr(o: Any) -> str:
            return re.sub('0x[0-9a-fA-F]+', '(mangled)', repr(o))

        self.assertEqual(
            mangle_repr(self.osr.associated_images),
            '<_AssociatedImageMap %s>' % mangle_repr(dict(self.osr.associated_images)),
        )

    def test_color_profile(self) -> None:
        self.assertIsNone(self.osr.color_profile)
        self.assertNotIn(
            'icc_profile', self.osr.read_region((0, 0), 0, (100, 100)).info
        )
        self.assertNotIn('icc_profile', self.osr.associated_images['thumbnail'].info)
        self.assertNotIn('icc_profile', self.osr.get_thumbnail((100, 100)).info)


# Requires DICOM support in OpenSlide.  Use associated image ICC support as
# a proxy.
@unittest.skipUnless(
    lowlevel.read_associated_image_icc_profile.available, "requires OpenSlide 4.0.0"
)
class TestDicomSlide(_Abstract.SlideTest):
    FILENAME = 'boxes_0.dcm'

    def test_color_profile(self) -> None:
        assert self.osr.color_profile is not None  # for type inference
        self.assertEqual(self.osr.color_profile.profile.device_class, 'mntr')
        main_profile = self.osr.read_region((0, 0), 0, (100, 100)).info['icc_profile']
        associated_profile = self.osr.associated_images['thumbnail'].info['icc_profile']
        self.assertEqual(len(main_profile), 456)
        self.assertEqual(main_profile, associated_profile)
        self.assertIs(main_profile, associated_profile)


class TestUnreadableSlide(_Abstract.SlideTest):
    FILENAME = 'unreadable.svs'

    def test_read_bad_region(self) -> None:
        self.assertEqual(self.osr.properties['openslide.vendor'], 'aperio')
        self.assertRaises(
            OpenSlideError, lambda: self.osr.read_region((0, 0), 0, (16, 16))
        )
        # verify that errors are sticky
        self.assertRaises(
            OpenSlideError, lambda: self.osr.properties['openslide.vendor']
        )

    def test_read_bad_associated_image(self) -> None:
        self.assertEqual(self.osr.properties['openslide.vendor'], 'aperio')
        # Prints "JPEGLib: Bogus marker length." to stderr due to
        # https://github.com/openslide/openslide/issues/36
        self.assertRaises(
            OpenSlideError, lambda: self.osr.associated_images['thumbnail']
        )
        # verify that errors are sticky
        self.assertRaises(
            OpenSlideError, lambda: self.osr.properties['openslide.vendor']
        )



#
# openslide-python - Python bindings for the OpenSlide library
#
# Copyright (c) 2016 Benjamin Gilbert
#
# This library is free software; you can redistribute it and/or modify it
# under the terms of version 2.1 of the GNU Lesser General Public License
# as published by the Free Software Foundation.
#
# This library is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public
# License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this library.  If not, see <https://www.gnu.org/licenses/>.
#

from __future__ import annotations

import unittest

from common import file_path

from openslide import ImageSlide, OpenSlide, lowlevel
from openslide.deepzoom import DeepZoomGenerator


class _Abstract:
    # nested class to prevent the test runner from finding it
    class BoxesDeepZoomTest(unittest.TestCase):
        CLASS: type | None = None
        FILENAME: str | None = None

        def setUp(self) -> None:
            assert self.CLASS is not None
            assert self.FILENAME is not None
            self.osr = self.CLASS(file_path(self.FILENAME))
            self.dz = DeepZoomGenerator(self.osr, 254, 1)

        def tearDown(self) -> None:
            self.osr.close()

        def test_repr(self) -> None:
            self.assertEqual(
                repr(self.dz),
                'DeepZoomGenerator(%r, tile_size=254, overlap=1, limit_bounds=False)'
                % self.osr,
            )

        def test_metadata(self) -> None:
            self.assertEqual(self.dz.level_count, 10)
            self.assertEqual(self.dz.tile_count, 11)
            self.assertEqual(
                self.dz.level_tiles,
                (
                    (1, 1),
                    (1, 1),
                    (1, 1),
                    (1, 1),
                    (1, 1),
                    (1, 1),
                    (1, 1),
                    (1, 1),
                    (1, 1),
                    (2, 1),
                ),
            )
            self.assertEqual(
                self.dz.level_dimensions,
                (
                    (1, 1),
                    (2, 1),
                    (3, 2),
                    (5, 4),
                    (10, 8),
                    (19, 16),
                    (38, 32),
                    (75, 63),
                    (150, 125),
                    (300, 250),
                ),
            )

        def test_get_tile(self) -> None:
            self.assertEqual(self.dz.get_tile(9, (1, 0)).size, (47, 250))

        def test_tile_color_profile(self) -> None:
            if self.CLASS is OpenSlide and not lowlevel.read_icc_profile.available:
                self.skipTest("requires OpenSlide 4.0.0")
            self.assertEqual(len(self.dz.get_tile(9, (1, 0)).info['icc_profile']), 588)

        def test_get_tile_bad_level(self) -> None:
            self.assertRaises(ValueError, lambda: self.dz.get_tile(-1, (0, 0)))
            self.assertRaises(ValueError, lambda: self.dz.get_tile(10, (0, 0)))

        def test_get_tile_bad_address(self) -> None:
            self.assertRaises(ValueError, lambda: self.dz.get_tile(0, (-1, 0)))
            self.assertRaises(ValueError, lambda: self.dz.get_tile(0, (1, 0)))

        def test_get_tile_coordinates(self) -> None:
            self.assertEqual(
                self.dz.get_tile_coordinates(9, (1, 0)), ((253, 0), 0, (47, 250))
            )

        def test_get_tile_dimensions(self) -> None:
            self.assertEqual(self.dz.get_tile_dimensions(9, (1, 0)), (47, 250))

        def test_get_dzi(self) -> None:
            self.assertTrue(
                'http://schemas.microsoft.com/deepzoom/2008' in self.dz.get_dzi('jpeg')
            )


class TestSlideDeepZoom(_Abstract.BoxesDeepZoomTest):
    CLASS = OpenSlide
    FILENAME = 'boxes.tiff'


class TestImageDeepZoom(_Abstract.BoxesDeepZoomTest):
    CLASS = ImageSlide
    FILENAME = 'boxes.png'



#
# openslide-python - Python bindings for the OpenSlide library
#
# Copyright (c) 2024 Benjamin Gilbert
#
# This library is free software; you can redistribute it and/or modify it
# under the terms of version 2.1 of the GNU Lesser General Public License
# as published by the Free Software Foundation.
#
# This library is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public
# License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this library.  If not, see <https://www.gnu.org/licenses/>.
#

from typing import Protocol

class _Buffer(Protocol):
    # Python 3.12+ has collections.abc.Buffer
    def __buffer__(self, flags: int) -> memoryview: ...

def argb2rgba(buf: _Buffer) -> None: ...



from pathlib import Path
import sys
import sysconfig

from setuptools import Extension, setup

# Load version string
with open(Path(__file__).parent / 'openslide/_version.py') as _fh:
    exec(_fh.read())  # instantiates __version__

# use the Limited API on Python 3.11+ on GIL builds; build release-specific
# wheels on older or free-threaded Python
_abi3 = sys.version_info >= (3, 11) and not sysconfig.get_config_var('Py_GIL_DISABLED')

setup(
    ext_modules=[
        Extension(
            'openslide._convert',
            ['openslide/_convert.c'],
            # hide symbols that aren't in the Limited API
            define_macros=[('Py_LIMITED_API', '0x030b0000')] if _abi3 else [],
            # tag extension module for Limited API
            py_limited_api=_abi3,
        ),
    ],
    options={
        # tag wheel for Limited API
        'bdist_wheel': {'py_limited_api': 'cp311'} if _abi3 else {},
    },
    #
    # setuptools < 61 compatibility for distro packages building from source
    name='openslide-python',
    version=__version__,  # type: ignore[name-defined]  # noqa: F821
    install_requires=[
        'Pillow',
    ],
    packages=[
        'openslide',
    ],
    package_data={
        'openslide': ['py.typed', '*.pyi'],
    },
    zip_safe=False,
)
